"""Client-side gRPC interceptors."""

import abc
from collections import namedtuple
import logging

from grpc import (
    ClientCallDetails,
    StatusCode,
    StreamStreamClientInterceptor,
    UnaryStreamClientInterceptor,
    UnaryUnaryClientInterceptor,
)

from ansys.edb.core.inner.exceptions import EDBSessionException, ErrorCode, InvalidArgumentException
from ansys.edb.core.inner.rpc_info_utils import can_cache
from ansys.edb.core.utility.io_manager import ServerNotification, get_io_manager


class Interceptor(
    UnaryUnaryClientInterceptor,
    UnaryStreamClientInterceptor,
    StreamStreamClientInterceptor,
    metaclass=abc.ABCMeta,
):
    """Provides the base interceptor class."""

    def __init__(self, logger):
        """Initialize a base interceptor with a logger."""
        super().__init__()
        self._logger = logger

    @abc.abstractmethod
    def _post_process(self, response):
        pass

    def _continue_unary_unary(self, continuation, client_call_details, request):
        return continuation(client_call_details, request)

    def intercept_unary_unary(self, continuation, client_call_details, request):
        """Intercept a gRPC call."""
        response = self._continue_unary_unary(continuation, client_call_details, request)

        self._post_process(response)

        return response

    def intercept_unary_stream(self, continuation, client_call_details, request):
        """Intercept a gRPC streaming call."""
        return continuation(client_call_details, request)

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        """Intercept a gRPC streaming call."""
        return continuation(client_call_details, request_iterator)


class LoggingInterceptor(Interceptor):
    """Logs EDB errors on each request."""

    _LOG_LEVELS = {
        "log-fatal": logging.CRITICAL,
        "log-error": logging.ERROR,
        "log-warn": logging.WARN,
        "log-info": logging.INFO,
    }

    def _handler(self, log_level):
        if log_level == logging.CRITICAL:
            return self._logger.critical
        elif log_level == logging.WARN:
            return self._logger.warning
        elif log_level == logging.ERROR:
            return self._logger.error
        elif log_level == logging.INFO:
            return self._logger.info
        else:
            return self._logger.logger

    def _post_process(self, response):
        tm = response.trailing_metadata()
        if tm is not None:
            for kv in tm:
                if kv[0] in self._LOG_LEVELS:
                    handler = self._handler(self._LOG_LEVELS[kv[0]])
                    handler(kv[1])


class ExceptionInterceptor(Interceptor):
    """Handles general gRPC errors on each request."""

    def _post_process(self, response):
        code = response.code()
        exception = None

        if code == StatusCode.UNAVAILABLE:
            exception = EDBSessionException(ErrorCode.UNAVAILABLE)
            self._logger.critical(repr(exception))
        elif code == StatusCode.INVALID_ARGUMENT:
            exception = InvalidArgumentException(response)
            self._logger.error(repr(exception))

        if exception is not None:
            raise exception


class IOInterceptor(Interceptor):
    """Returns cached values if a given request has already been made and caching is enabled."""

    def __init__(self, logger, rpc_counter):
        """Initialize a caching interceptor with a logger and rpc counter."""
        super().__init__(logger)
        self._rpc_counter = rpc_counter
        self._reset_cache_entry_data()

    def _reset_cache_entry_data(self):
        self._current_rpc_method = ""
        self._current_cache_key_details = None
        self._hijacked = False

    def _should_log_traffic(self):
        return self._rpc_counter is not None

    class _ClientCallDetails(
        namedtuple("_ClientCallDetails", ("method", "timeout", "metadata", "credentials")),
        ClientCallDetails,
    ):
        pass

    @staticmethod
    def _add_caching_option_to_metadata(metadata, option, is_enabled):
        metadata.append((option, "1" if is_enabled else "0"))

    @classmethod
    def _get_client_call_details_with_caching_options(cls, client_call_details):
        io_mgr = get_io_manager()
        if not io_mgr.get_notifications_for_server():
            return client_call_details
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        for notification in io_mgr.get_notifications_for_server(True):
            if notification == ServerNotification.INVALIDATE_CACHE:
                cls._add_caching_option_to_metadata(metadata, "invalidate-cache", True)
            elif notification == ServerNotification.FLUSH_BUFFER:
                cls._add_caching_option_to_metadata(metadata, "flush-buffer", True)
            elif notification == ServerNotification.RESET_FUTURE_TRACKING:
                cls._add_caching_option_to_metadata(metadata, "reset-future-tracking", True)
        return cls._ClientCallDetails(
            client_call_details.method,
            client_call_details.timeout,
            metadata,
            client_call_details.credentials,
        )

    @staticmethod
    def _attempt_hijack(*args):
        io_manager = get_io_manager()
        hijacked_response = None
        if (buffer := io_manager.buffer) is not None:
            hijacked_response = buffer.hijack_request(*args)
        if hijacked_response is None and (cache := io_manager.cache) is not None:
            hijacked_response = cache.hijack_request(*args)
        return hijacked_response

    def _hijack(self, client_call_details, request):
        io_manager = get_io_manager()
        if io_manager.is_enabled and not io_manager.is_blocking:
            with io_manager.manage_io():
                method_tokens = client_call_details.method.strip("/").split("/")
                cache_key_details = method_tokens[0], method_tokens[1], request
                if (hijacked_result := self._attempt_hijack(*cache_key_details)) is not None:
                    self._hijacked = True
                    return hijacked_result
                if io_manager.cache is not None and can_cache(
                    cache_key_details[0], cache_key_details[1]
                ):
                    self._current_cache_key_details = cache_key_details
                io_manager.add_notification_for_server(ServerNotification.RESET_FUTURE_TRACKING)
        if self._should_log_traffic():
            self._current_rpc_method = client_call_details.method

    def _continue_unary_unary(self, continuation, client_call_details, request):
        if (hijacked_result := self._hijack(client_call_details, request)) is not None:
            return hijacked_result
        return super()._continue_unary_unary(
            continuation,
            self._get_client_call_details_with_caching_options(client_call_details),
            request,
        )

    def _post_process(self, response):
        io_manager = get_io_manager()
        if (cache := io_manager.cache) is not None and self._current_cache_key_details is not None:
            cache.add(*self._current_cache_key_details, response.result())
        if self._should_log_traffic() and not self._hijacked:
            self._rpc_counter[self._current_rpc_method] += 1
        self._reset_cache_entry_data()

    def intercept_unary_stream(self, continuation, client_call_details, request):
        """Intercept a gRPC streaming call."""
        if (hijacked_result := self._hijack(client_call_details, request)) is not None:
            return hijacked_result
        return super().intercept_unary_stream(
            continuation,
            self._get_client_call_details_with_caching_options(client_call_details),
            request,
        )

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        """Intercept a gRPC streaming call."""
        if (hijacked_result := self._hijack(client_call_details, request_iterator)) is not None:
            return hijacked_result
        return super().intercept_stream_stream(
            continuation,
            self._get_client_call_details_with_caching_options(client_call_details),
            request_iterator,
        )



"""Padstack definition."""

from ansys.api.edb.v1 import padstack_def_pb2_grpc
import ansys.api.edb.v1.padstack_def_pb2 as pb

from ansys.edb.core.definition.padstack_def_data import PadstackDefData
from ansys.edb.core.edb_defs import DefinitionObjType
from ansys.edb.core.inner import ObjBase
from ansys.edb.core.session import StubAccessor, StubType


class PadstackDef(ObjBase):
    """Represents a padstack definition."""

    __stub: padstack_def_pb2_grpc.PadstackDefServiceStub = StubAccessor(StubType.padstack_def)

    @classmethod
    def create(cls, db, name):
        """Create a padstack definition in a given database.

        Parameters
        ----------
        db : :class:`.Database`
            Database to create the padstack definition in.
        name : str
            Data to set on the padstack definition.

        Returns
        -------
        PadstackDef
            Padstack definition created.
        """
        return PadstackDef(cls.__stub.Create(PadstackDef._padstack_def_string_message(db, name)))

    @classmethod
    def find_by_name(cls, db, name):
        """Find a padstack definition by name in a given database.

        Parameters
        ----------
        db : :class:`.Database`.
            Database to search for the padstack definition.
        name : str
            Name of the padstack definition.

        Returns
        -------
        PadstackDef
            Padstack definition found.
        """
        return PadstackDef(
            cls.__stub.FindByName(PadstackDef._padstack_def_string_message(db, name))
        )

    @property
    def definition_type(self):
        """:class:`.DefinitionObjType`: Definition type."""
        return DefinitionObjType.PADSTACK_DEF

    @property
    def name(self):
        """:obj:`str`: Name of the padstack definition.

        This property is read-only.
        """
        return self.__stub.GetName(self.msg).value

    @property
    def data(self):
        """:class:`.PadstackDefData`: \
        Data for the padstack definition."""
        return PadstackDefData(self.__stub.GetData(self.msg))

    @data.setter
    def data(self, data):
        self.__stub.SetData(pb.PadstackDefSetDataMessage(target=self.msg, data=data.msg))

    def delete(self):
        """Delete the padstack definition."""
        self.__stub.Delete(self.msg)

    @staticmethod
    def _padstack_def_string_message(target, name):
        """Create a string message for the padstack definition.

        Parameters
        ----------
        target : :class:`.Database` or PadstackDef
        name : str
           Name of the string message.

        Returns
        -------
        PadstackDefStringMessage
        """
        return pb.PadstackDefStringMessage(target=target.msg, name=name)



"""Connection object."""
from ansys.api.edb.v1 import connectable_pb2

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.inner import layout_obj, messages
from ansys.edb.core.inner.factory import create_lyt_obj
from ansys.edb.core.layout import mcad_model as mm
from ansys.edb.core.session import ConnectableServiceStub, StubAccessor, StubType


class ConnObj(layout_obj.LayoutObj):
    """Provides the base class representing the connection object."""

    __stub: ConnectableServiceStub = StubAccessor(StubType.connectable)
    layout_obj_type = LayoutObjType.INVALID_LAYOUT_OBJ

    @classmethod
    def _validate_edb_obj_type(cls, edb_obj_msg):
        """Verify that the object type received from the server matches the object type requested by the client."""
        client_obj = cls(edb_obj_msg)
        if cls.layout_obj_type == LayoutObjType.PRIMITIVE:
            import ansys.edb.core.primitive.primitive as primitive

            def get_client_prim_type_from_class():
                if cls == primitive.Circle:
                    return primitive.PrimitiveType.CIRCLE
                if cls == primitive.Rectangle:
                    return primitive.PrimitiveType.RECTANGLE
                if cls == primitive.Polygon:
                    return primitive.PrimitiveType.POLYGON
                if cls == primitive.Path:
                    return primitive.PrimitiveType.PATH
                if cls == primitive.Bondwire:
                    return primitive.PrimitiveType.BONDWIRE
                if cls == primitive.BoardBendDef:
                    return primitive.PrimitiveType.BOARD_BEND
                if cls == primitive.Text:
                    return primitive.PrimitiveType.TEXT

            if get_client_prim_type_from_class() == primitive.Primitive(edb_obj_msg).primitive_type:
                return client_obj
        elif cls.layout_obj_type == LayoutObjType.TERMINAL:
            from ansys.edb.core.terminal import terminals

            server_term_type = terminals.Terminal(edb_obj_msg).type
            if client_obj.type == server_term_type:
                return client_obj
        else:
            return client_obj
        return cls(None)

    def cast(self):
        """Cast the ConnObj object to the correct concrete type.

        Returns
        -------
        .ConnObj
        """
        return create_lyt_obj(self.msg, self.obj_type)

    @property
    def obj_type(self):
        """:class:`.LayoutObjType`: Layout object type.

        This property is read-only.
        """
        if self.layout_obj_type != LayoutObjType.INVALID_LAYOUT_OBJ:
            return super().obj_type
        else:
            return LayoutObjType(self.__stub.GetObjType(self.msg).type)

    @classmethod
    def find_by_id(cls, layout, uid):
        """Find a :term:`Connectable` object by database ID in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the :term:`Connectable` object.
        uid : int
            Database ID.

        Returns
        -------
        :term:`Connectable`
            Connectable object with the given database ID.
        """
        found_edb_obj_msg = cls.__stub.FindByIdAndType(
            connectable_pb2.FindByIdLayoutObjMessage(
                layout=layout.msg,
                type=cls.layout_obj_type.value,
                id_msg=messages.edb_internal_id_message(uid),
            )
        )
        return cls._validate_edb_obj_type(found_edb_obj_msg)

    @property
    def edb_uid(self):
        """:obj:`int`: Unique, persistent ID for the :term:`Connectable` object.

        This ID is unique across all :term:`Connectable` objects in the cell and persistent across closing and \
        reopening the database.

        This property is read-only.
        """
        return self.__stub.GetId(self.msg).id

    @property
    def component(self):
        """:class:`.ComponentGroup`: \
        Component of the :term:`Connectable` object."""
        from ansys.edb.core.hierarchy.component_group import ComponentGroup

        return ComponentGroup(self.__stub.GetComponent(self.msg))

    @property
    def group(self):
        """:class:`.Group` object."""
        from ansys.edb.core.hierarchy.group import Group

        return Group(self.__stub.GetGroup(self.msg)).cast()

    @group.setter
    def group(self, group):
        self.__stub.SetGroup(messages.pointer_property_message(target=self, value=group))

    @property
    def net(self):
        """:class:`.Net`: Net of the :term:`Connectable` object.

        This property can be set with a :class:`.Net` instance, a string, or :obj:`None`.
        """
        from ansys.edb.core.net.net import Net

        return Net(self.__stub.GetNet(self.msg))

    @net.setter
    def net(self, net):
        self.__stub.SetNet(
            connectable_pb2.SetNetMessage(
                target=self.msg,
                net=messages.net_ref_message(net),
            )
        )

    def create_stride(self):
        """Create a Stride model from an MCAD file.

        Returns
        -------
        :class:`.McadModel`
            Stride model created.
        """
        return mm.McadModel.create_stride(connectable=self)

    def create_hfss(self):
        """Create an HFSS model from an MCAD file.

        Returns
        -------
        :class:`.McadModel`
            HFSS model created.
        """
        return mm.McadModel.create_hfss(connectable=self)

    def create_3d_comp(self):
        """Create a 3D composite model from an MCAD file.

        Returns
        -------
        :class:`.McadModel`
            3D composite model created.
        """
        return mm.McadModel.create_3d_comp(connectable=self)

    @property
    def is_mcad(self):
        """:obj:`bool`: Flag indicating if this is an MCAD model.

        This property is read-only.
        """
        return mm.McadModel.is_mcad(self)

    @property
    def is_mcad_stride(self):
        """:obj:`bool`: Flag indicating if this is a Stride MCAD model.

        This property is read-only.
        """
        return mm.McadModel.is_mcad_stride(self)

    @property
    def is_mcad_hfss(self):
        """:obj:`bool`: Flag indicating if this is an HFSS MCAD model.

        This property is read-only.
        """
        return mm.McadModel.is_mcad_hfss(self)

    @property
    def is_mcad_3d_comp(self):
        """:obj:`bool`: Flag indicating if this is a 3D composite MCAD model.

        This property is read-only.
        """
        return mm.McadModel.is_mcad_3d_comp(self)



"""Solder ball property."""

import ansys.api.edb.v1.padstack_def_data_pb2 as padstack_def_data_pb2
from ansys.api.edb.v1.padstack_def_data_pb2 import SolderballShape
import ansys.api.edb.v1.solder_ball_property_pb2 as solder_ball_property_pb2
from ansys.api.edb.v1.solder_ball_property_pb2_grpc import SolderBallPropertyServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.definition.padstack_def_data import SolderballPlacement, SolderballShape
from ansys.edb.core.inner import ObjBase
from ansys.edb.core.inner.messages import (
    edb_obj_message,
    string_property_message,
    value_message,
    value_property_message,
)
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class SolderBallProperty(ObjBase):
    """Represents a solder ball property."""

    __stub: SolderBallPropertyServiceStub = StubAccessor(StubType.solder_ball_property)

    @classmethod
    def create(cls):
        """
        Create a solder ball property.

        Returns
        -------
        SolderBallProperty
            Solder ball property created.
        """
        return SolderBallProperty(cls.__stub.Create(empty_pb2.Empty()))

    @property
    def material_name(self):
        """:obj:`str`: Name of the solder ball material."""
        return self.__stub.GetMaterialName(edb_obj_message(self)).value

    @material_name.setter
    def material_name(self, name):
        self.__stub.SetMaterialName(string_property_message(self, name))

    @property
    def height(self):
        """:term:`ValueLike`: Height of the solder ball."""
        return Value(self.__stub.GetHeight(edb_obj_message(self)))

    @height.setter
    def height(self, height):
        self.__stub.SetHeight(value_property_message(self, value_message(height)))

    def get_diameter(self):
        """Get the diameter parameters of the solder ball property.

        Returns
        -------
        diameter1 : :term:`ValueLike`
            Diameter 1 of the solder ball property.
        diameter2 : :term:`ValueLike`
            Diameter 2 of the solder ball property.
        """
        diameter_paramaters = self.__stub.GetDiameter(edb_obj_message(self))
        return Value(diameter_paramaters.diameter1), Value(diameter_paramaters.diameter2)

    def set_diameter(self, diameter1, diameter2):
        """Set the diameter parameters of the solder ball property.

        Parameters
        ----------
        diameter1 : :term:`ValueLike`
            Diameter 1 of the solder ball property.
        diameter2 : :term:`ValueLike`
            Diameter 2 of the solder ball property.
        """
        self.__stub.SetDiameter(
            solder_ball_property_pb2.SetDiameterMessage(
                target=edb_obj_message(self),
                value=solder_ball_property_pb2.DiameterMessage(
                    diameter1=value_message(diameter1),
                    diameter2=value_message(diameter2),
                ),
            )
        )

    @property
    def uses_solderball(self):
        """:obj:`bool`: Flag indicating if the solder ball is used.

        This property is read-only.
        """
        return self.__stub.UsesSolderball(self.msg).value

    @property
    def shape(self):
        """:class:`SolderballShape`: Solder ball shape."""
        return SolderballShape(self.__stub.GetShape(self.msg).solderball_shape)

    @shape.setter
    def shape(self, shape):
        self.__stub.SetShape(
            padstack_def_data_pb2.PadstackDefDataSetSolderballShapeMessage(
                target=self.msg, solderball_shape=shape.value
            )
        )

    @property
    def placement(self):
        """:class:`SolderballPlacement`: Solder ball placement."""
        return SolderballPlacement(self.__stub.GetPlacement(self.msg).solderball_placement)

    @placement.setter
    def placement(self, placement):
        self.__stub.SetPlacement(
            padstack_def_data_pb2.PadstackDefDataSetSolderballPlacementMessage(
                target=self.msg, solderball_placement=placement.value
            )
        )

    def clone(self):
        """
        Clone a solder ball property.

        Returns
        -------
        SolderBallProperty
            Solder ball property cloned.
        """
        return SolderBallProperty(self.__stub.Clone(edb_obj_message(self)))



"""Layer collection."""

from enum import Enum

import ansys.api.edb.v1.layer_collection_pb2 as layer_collection_pb2

from ansys.edb.core.inner import ObjBase
from ansys.edb.core.inner.messages import (
    get_product_property_ids_message,
    get_product_property_message,
    set_product_property_message,
)
from ansys.edb.core.layer.layer import Layer, LayerType
from ansys.edb.core.layer.stackup_layer import StackupLayer
from ansys.edb.core.session import get_layer_collection_stub


class LayerCollectionMode(Enum):
    """Provides an enum representing possible modes of the layer collection."""

    LAMINATE = layer_collection_pb2.LAMINATE
    OVERLAPPING = layer_collection_pb2.OVERLAPPING
    MULTIZONE = layer_collection_pb2.MULTIZONE


class LayerTypeSet(Enum):
    """Provides an enum representing layer type sets used for filtering layers."""

    STACKUP_LAYER_SET = 0
    SIGNAL_LAYER_SET = 1
    DIELECTRIC_LAYER_SET = 2
    NON_STACKUP_LAYER_SET = 3
    ALL_LAYER_SET = 4


class DielectricMergingMethod(Enum):
    """Provides an enum representing dielectric merging method options."""

    WEIGHTED_AVERAGE = layer_collection_pb2.WEIGHTED_AVERAGE
    KRASZEWSKI = layer_collection_pb2.KRASZEWSKI
    WEIGHTED_CAPACITANCE = layer_collection_pb2.WEIGHTED_CAPACITANCE


def _layer_collection_zone_message(layer_collection, zone):
    """Convert to a ``LayerCollectionZoneMessage`` object."""
    return layer_collection_pb2.LayerCollectionZoneMessage(
        layer_collection=layer_collection.msg, zone=zone
    )


class LayerCollection(ObjBase):
    """Represents a layer collection."""

    @staticmethod
    def create(mode=LayerCollectionMode.LAMINATE):
        """Create a layer collection.

        Parameters
        ----------
        mode : LayerCollectionMode, default: LAMINATE
            Mode of the layer collection.

        Returns
        -------
        LayerCollection
            Layer collection created.
        """
        return LayerCollection(
            get_layer_collection_stub().Create(
                layer_collection_pb2.LayerCollectionModeMessage(mode=mode.value)
            )
        )

    def clone(self):
        """Create a clone of the layer collection.

        Returns
        -------
        LayerCollection
           Layer collection cloned.
        """
        return LayerCollection(get_layer_collection_stub().Clone(self.msg))

    @property
    def mode(self):
        """:class:`LayerCollectionMode`: Mode of the layer collection."""
        return LayerCollectionMode(get_layer_collection_stub().GetMode(self.msg).mode)

    @mode.setter
    def mode(self, mode):
        get_layer_collection_stub().SetMode(
            layer_collection_pb2.SetLayerCollectionModeMessage(
                layer_collection=self.msg, mode=mode.value
            )
        )

    def add_layers(self, layers):
        """Add a list of layers to the layer collection.

        Parameters
        ----------
        layers : list[Layer]
           List of layers.
        """
        layer_msgs = [lyr.msg for lyr in layers]
        get_layer_collection_stub().AddLayers(
            layer_collection_pb2.AddLayersMessage(layer_collection=self.msg, layers=layer_msgs)
        )

    def import_from_control_file(self, control_file_path, schema_file_path=None):
        """Import layers from a control file and optional XML schema file.

        Parameters
        ----------
        control_file_path : str
            Full path to the control file.
        schema_file_path : str, default: None
            Full path to the XML schema file.
        """
        import_msg = layer_collection_pb2.ImportFromControlFileMessage(
            layer_collection=self.msg, control_file_path=control_file_path
        )
        if schema_file_path is not None:
            import_msg.schema_path = schema_file_path
        get_layer_collection_stub().ImportFromControlFile(import_msg)

    def _add_layer(self, layer, above_below=None, add_top=None):
        """Pack layer addition arguments into the ``AddLayersMessage`` object and send to the server."""
        add_layer_msg = layer_collection_pb2.AddLayerMessage(
            layer_collection=self.msg, layer=layer.msg
        )
        if above_below is not None:
            above_below_msg = layer_collection_pb2.AddLayerAboveBelowMessage(
                above_below_layer_name=above_below[0], add_above=above_below[1]
            )
            add_layer_msg.above_below_msg.CopyFrom(above_below_msg)
        elif add_top is not None:
            add_layer_msg.add_top = add_top
        return Layer(get_layer_collection_stub().AddLayer(add_layer_msg)).cast()

    def _add_layer_relative(self, layer, relative_layer_name, add_above):
        """Add a layer above or below another layer."""
        return self._add_layer(layer, (relative_layer_name, add_above))

    def add_layer_above(self, layer_to_add, layer_to_add_above_name):
        """Add a new layer above the specified layer.

        This method adjusts existing layers as needed to maintain stackup consistency.

        Parameters
        ----------
        layer_to_add : Layer
            Name of the layer to add.
        layer_to_add_above_name : str
            Name of the layer above which to add the new layer.

        Returns
        -------
        Layer
        """
        return self._add_layer_relative(layer_to_add, layer_to_add_above_name, True)

    def add_layer_below(self, layer_to_add, layer_to_add_below_name):
        """Add a new layer below the specified layer.

        This method adjusts existing layers as needed to maintain stackup consistency.

        Parameters
        ----------
        layer_to_add : Layer
            Name of the layer to add.
        layer_to_add_below_name : str
            Name of the layer below which to add the new layer.

        Returns
        -------
        Layer
        """
        return self._add_layer_relative(layer_to_add, layer_to_add_below_name, False)

    def add_layer_top(self, layer_to_add):
        """Add a new layer to the top of the layer collection.

        This method adjusts existing layers as needed to maintain stackup consistency.

        Parameters
        ----------
        layer_to_add : Layer
            Name of the layer to add.

        Returns
        -------
        Layer
        """
        return self._add_layer(layer_to_add, add_top=True)

    def add_layer_bottom(self, layer_to_add):
        """Add a new layer to the bottom of the layer collection.

        This method adjusts existing layers as needed to maintain stackup consistency.

        Parameters
        ----------
        layer_to_add : Layer
            Name of the layer to add.

        Returns
        -------
        Layer
        """
        return self._add_layer(layer_to_add, add_top=False)

    def add_stackup_layer_at_elevation(self, stackup_layer_to_add):
        """Add a stackup layer at a user-specified elevation.

        This method doe not change the elevenation of other stackup layers.

        Parameters
        ----------
        stackup_layer_to_add : StackupLayer

        Returns
        -------
        StackupLayer
        """
        return self._add_layer(stackup_layer_to_add)

    def add_via_layer(self, via_layer_to_add):
        """Add a via layer to the layer collection.

        Parameters
        ----------
        via_layer_to_add : ViaLayer

        Returns
        -------
        ViaLayer
        """
        return self._add_layer(via_layer_to_add)

    def is_valid(self):
        """Determine if the layer collection is in a valid state.

        For a laminate stackup, this method checks whether there is layer overlapping or a gap.
        For an overlapping stackup, this method checks whether there is a dielectric layer
        overlapping or a gap.

        Returns
        -------
        bool
            ``True`` if the layer collection is in a valid state, ``False`` otherwise.
        """
        return get_layer_collection_stub().IsValid(self.msg).value

    def find_by_name(self, layer_name):
        """Find a layer in the layer collection.

        Parameters
        ----------
        layer_name : str
           Layer name.

        Returns
        -------
        Layer
        """
        return Layer(
            get_layer_collection_stub().FindByName(
                layer_collection_pb2.FindLayerByNameMessage(
                    layer_collection=self.msg, name=layer_name
                )
            )
        ).cast()

    @staticmethod
    def _get_layer_filter(layer_types):
        """Convert a list of layer types to an integer representation of a layer filter."""
        if not isinstance(layer_types, list):
            layer_types = [layer_types]
        layer_filter = 0
        for layer_type in layer_types:
            layer_type_value = layer_type.value
            if layer_type_value >= 0:
                layer_filter = layer_filter | (1 << layer_type_value)
        return layer_filter

    @staticmethod
    def _get_layer_filter_from_layer_type_set(layer_type_set):
        def _get_layer_type_list():
            if layer_type_set == LayerTypeSet.STACKUP_LAYER_SET:
                return [
                    LayerType.DIELECTRIC_LAYER,
                    LayerType.CONDUCTING_LAYER,
                    LayerType.SIGNAL_LAYER,
                ]
            elif layer_type_set == LayerTypeSet.SIGNAL_LAYER_SET:
                return [LayerType.CONDUCTING_LAYER, LayerType.SIGNAL_LAYER]
            elif layer_type_set == LayerTypeSet.DIELECTRIC_LAYER_SET:
                return [LayerType.DIELECTRIC_LAYER]
            elif layer_type_set == LayerTypeSet.NON_STACKUP_LAYER_SET:
                return list(LayerType)[LayerType.AIRLINES_LAYER.value : -2]
            else:
                return []

        return LayerCollection._get_layer_filter(_get_layer_type_list())

    def get_top_bottom_stackup_layers(self, layer_type_set):
        """Get the top and bottom stackup layers of a specific type and their elevations.

        Parameters
        ----------
        layer_type_set : LayerTypeSet
            Layer type set indicating the layer types to retrieve.

        Returns
        -------
        tuple[Layer, float, Layer, float]
            Returns a tuple in this format:
            (upper_layer, upper_layer_top_elevation, lower_layer, lower_layer_lower_elevation)
        """
        request = layer_collection_pb2.GetTopBottomStackupLayersMessage(
            layer_collection=self.msg,
            layer_type_set=LayerCollection._get_layer_filter_from_layer_type_set(layer_type_set),
        )
        response = get_layer_collection_stub().GetTopBottomStackupLayers(request)
        return (
            Layer(response.top_layer).cast(),
            response.top_layer_elevation,
            Layer(response.bottom_layer).cast(),
            response.bottom_layer_elevation,
        )

    def get_layers(self, layer_filter=LayerTypeSet.ALL_LAYER_SET):
        """Get a list of layers in the layer collection using a layer filter.

        Parameters
        ----------
        layer_filter : LayerTypeSet or LayerType or list[LayerType], default: ALL_LAYER_SET
            Layer filter.

        Returns
        -------
        list[Layer]
            List of layers based on the filter used.
        """
        layer_filter_int = (
            LayerCollection._get_layer_filter_from_layer_type_set(layer_filter)
            if isinstance(layer_filter, LayerTypeSet)
            else LayerCollection._get_layer_filter(layer_filter)
        )

        response = get_layer_collection_stub().GetLayers(
            layer_collection_pb2.GetLayersMessage(
                layer_collection=self.msg, layer_filter=layer_filter_int
            )
        )

        return [Layer(msg).cast() for msg in response.items]

    def get_product_property(self, prod_id, attr_it):
        """Get the product property of the layer collection for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            Product ID.
        attr_it : int
            Attribute ID.

        Returns
        -------
        str
            Product property.
        """
        return (
            get_layer_collection_stub()
            .GetProductProperty(get_product_property_message(self, prod_id, attr_it))
            .value
        )

    def set_product_property(self, prod_id, attr_it, prop_value):
        """Set the product property of the layer collection for a given product ID and attribute ID.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            Product ID.
        attr_it : int
            Attribute ID.
        prop_value : str
            New property value.
        """
        get_layer_collection_stub().SetProductProperty(
            set_product_property_message(self, prod_id, attr_it, prop_value)
        )

    def get_product_property_ids(self, prod_id):
        """Get a list of attribute IDs for a given product ID for the layer collection.

        Parameters
        ----------
        prod_id : :class:`.ProductIdType`
            Product ID.

        Returns
        -------
        list[int]
            List of attribute IDs for the given product ID.
        """
        attr_ids = (
            get_layer_collection_stub()
            .GetProductPropertyIds(get_product_property_ids_message(self, prod_id))
            .ids
        )
        return [attr_id for attr_id in attr_ids]

    def merge_dielectrics(
        self,
        layout,
        start_layer_name,
        end_layer_name,
        merging_method,
        merged_layer_name,
        merged_mat_name,
    ):
        """Merge the dielectric layers in a range of layers into one large dielectric layer.

        Parameters
        ----------
        layout : :class:`.Layout`
        start_layer_name : str
        end_layer_name : str
        merging_method : DielectricMergingMethod
        merged_layer_name : str
        merged_mat_name : str

        Returns
        -------
        StackupLayer
        """
        return StackupLayer(
            get_layer_collection_stub().MergeDielectrics(
                layer_collection_pb2.MergeDielectricsMessage(
                    layer_collection=self.msg,
                    layout=layout.msg,
                    start_layer_name=start_layer_name,
                    end_layer_name=end_layer_name,
                    merging_method=merging_method.value,
                    merged_layer_name=merged_layer_name,
                    merged_mat_name=merged_mat_name,
                )
            )
        )

    @property
    def zone_ids(self):
        r""":obj:`list`\[:obj:`int`\]: List of all zones in the layer collection."""
        zones = get_layer_collection_stub().GetZoneIds(self.msg).zones
        return [zone for zone in zones]

    def get_zone_name(self, zone):
        """Get the name for a given zone.

        Parameters
        ----------
        zone : int
           Zone ID.

        Returns
        -------
        str
            Name of the zone.
        """
        return (
            get_layer_collection_stub()
            .GetZoneName(_layer_collection_zone_message(self, zone))
            .value
        )

    def set_zone_name(self, zone, name):
        """Set the name for a given zone.

        Parameters
        ----------
        zone : int
            Zone ID.
        name : str
            New name to give the zone.
        """
        request = layer_collection_pb2.SetZoneNameMessage(
            layer_collection=self.msg, zone=zone, zone_name=name
        )
        get_layer_collection_stub().SetZoneName(request)

    def insert_zone(self, copy_zone=-1):
        """Insert a zone.

        Parameters
        ----------
        copy_zone : int, default: -1
            Zone to copy from when inserting a new zone.
            If valid, the new zone is inserted as a copy of the given zone.
            Otherwise, the new zone is empty.

        Returns
        -------
        int
            ID of the zone inserted if successful.
        """
        request = layer_collection_pb2.InsertZoneMessage(
            layer_collection=self.msg, copy_zone=copy_zone
        )
        return get_layer_collection_stub().InsertZone(request).value

    def remove_zone(self, zone):
        """Remove a zone.

        Parameters
        ----------
        zone : int
           ID of the zone.
        """
        get_layer_collection_stub().RemoveZone(_layer_collection_zone_message(self, zone))

    def simplify_dielectrics_for_phi(
        self,
        database,
        layer_thickness_thresh=-1,
        merging_method=DielectricMergingMethod.WEIGHTED_CAPACITANCE,
    ):
        """Split dielectric layers at the boundaries of signal layers and merge them.

        Parameters
        ----------
        database : :class:`.Database`
        layer_thickness_thresh : float, default: -1
           Thickness threshold for the layer.
        merging_method : DielectricMergingMethod, default: WEIGHTED_CAPACITANCE
           Method for merging.

        Returns
        -------
        list[StackupLayer]
            List of dielectric layers created during the dielectric simplification process.
        """
        simplified_lyrs = (
            get_layer_collection_stub()
            .SimplifyDielectricsForPhi(
                layer_collection_pb2.SimplifyDielectricsForPhiMessage(
                    layer_collection=self.msg,
                    database=database.msg,
                    layer_thickness_thresh=layer_thickness_thresh,
                    merging_method=merging_method.value,
                )
            )
            .items
        )
        return [StackupLayer(simplified_lyr) for simplified_lyr in simplified_lyrs]

    def add_zone_to_layer(self, layer, zone, in_zone):
        """Set the zone to the layer and update the layer in the collection.

        Parameters
        ----------
        layer : Layer
        zone : int
        in_zone : bool
        """
        request = layer_collection_pb2.AddZoneToLayerMessage(
            layer_collection=self.msg, layer=layer.msg, zone=zone, in_zone=in_zone
        )
        get_layer_collection_stub().AddZoneToLayer(request)



"""Q3D simulation setup."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ansys.edb.core.simulation_setup.q3d_simulation_settings import Q3DSimulationSettings
from ansys.edb.core.simulation_setup.simulation_setup import SimulationSetup, SimulationSetupType

if TYPE_CHECKING:
    from ansys.edb.core.layout.cell import Cell


class Q3DSimulationSetup(SimulationSetup):
    """Represents Q3D simulation setup data."""

    @classmethod
    def create(cls, cell: Cell, name: str) -> Q3DSimulationSetup:
        """Create a Q3D simulation setup.

        Parameters
        ----------
        cell : :class:`.Cell`
            Cell to create the simulation setup in.
        name : str
            Name of the simulation setup.

        Returns
        -------
        Q3DSimulationSetup
            Q3D simulation setup created.
        """
        return super()._create(cell, name, SimulationSetupType.Q3D_SIM)

    @property
    def settings(self) -> Q3DSimulationSettings:
        """:class:`.Q3DSimulationSettings`: Simulation settings of the Q3D simulation setup."""
        return Q3DSimulationSettings(self)



"""RLC component property."""

from ansys.api.edb.v1.rlc_component_property_pb2_grpc import RLCComponentPropertyServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.definition.component_property import ComponentProperty
from ansys.edb.core.inner import messages
from ansys.edb.core.session import StubAccessor, StubType


class RLCComponentProperty(ComponentProperty):
    """Represents an RLC component property."""

    __stub: RLCComponentPropertyServiceStub = StubAccessor(StubType.rlc_component_property)

    @classmethod
    def create(cls):
        """
        Create an RLC component property.

        Returns
        -------
        RLCComponentProperty
            RLC component property created.
        """
        return RLCComponentProperty(cls.__stub.Create(empty_pb2.Empty()))

    @property
    def enabled(self):
        """:obj:`bool`: Flag indicating if the RLC component property is enabled."""
        return self.__stub.GetEnabled(messages.edb_obj_message(self)).value

    @enabled.setter
    def enabled(self, value):
        self.__stub.SetEnabled(messages.bool_property_message(self, value))



"""Extended net."""

import ansys.api.edb.v1.extended_net_pb2 as enet_pb2

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.net import net_class
from ansys.edb.core.session import StubAccessor, StubType


class ExtendedNet(net_class.NetClass):
    """Represents an extended net."""

    __stub = StubAccessor(StubType.extended_net)
    layout_obj_type = LayoutObjType.EXTENDED_NET

    @classmethod
    def create(cls, layout, name):
        """Create an extended net.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the extended net in.
        name : str
            Name of the extended net.

        Returns
        -------
        ExtendedNet
            Extended net created.
        """
        return ExtendedNet(
            cls.__stub.Create(enet_pb2.ExtendedNetCreationMessage(layout=layout.msg, name=name))
        )

    @classmethod
    def find_by_name(cls, layout, name):
        """Find an extended net by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the extended net.
        name : str
            Name of the extended net.

        Returns
        -------
        ExtendedNet
            Extended net that was found. Check the :obj:`is_null <.ExtendedNet.is_null>`
            property of the extended net to see if it exists.
        """
        return ExtendedNet(
            cls.__stub.FindByName(enet_pb2.ExtendedNetLookupMessage(layout=layout.msg, name=name))
        )

    def add_net(self, net):
        """Add a net to the extended net.

        Parameters
        ----------
        net : Net
            Net to add.
        """
        self.__stub.AddNet(ExtendedNet._extnet_modify_net_msg(self, net))

    def remove_net(self, net):
        """Remove a net from the extended net.

        Parameters
        ----------
        net : Net
            Net to remove.
        """
        self.__stub.RemoveNet(ExtendedNet._extnet_modify_net_msg(self, net))

    def remove_all_nets(self):
        """Remove all nets from the extended net."""
        self.__stub.RemoveAllNets(self.msg)

    @staticmethod
    def _extnet_modify_net_msg(ext_net, net):
        return enet_pb2.ExtendedNetModifyMessage(ext_net=ext_net.msg, net=net.msg)



"""Terminal Instance Terminal."""

from ansys.edb.core.inner import TypeField, messages
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.terminal.terminal import Terminal, TerminalType
from ansys.edb.core.terminal.terminal_instance import TerminalInstance


class TerminalInstanceTerminal(Terminal):
    """Represents a terminal instance terminal."""

    __stub = StubAccessor(StubType.terminal_instance_terminal)
    type = TypeField(TerminalType.TERM_INST)

    @classmethod
    def create(cls, layout, term_instance, name, net_ref, is_ref=False):
        """Create a terminal instance terminal.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the terminal instance terminal in.
        term_instance : TerminalInstance
            Terminal instance to create the terminal instance terminal on.
        name : :obj:`str`
            Name of the terminal instance terminal.
        net_ref : :class:`.Net` or :obj:`str` or None
            Net reference.
        is_ref : bool, default: False
            Whether the terminal instance terminal is a reference terminal.

        Returns
        -------
        TerminalInstanceTerminal
        """
        return TerminalInstanceTerminal(
            cls.__stub.Create(
                messages.term_inst_term_creation_message(
                    layout, net_ref, name, term_instance, is_ref
                )
            )
        )

    @property
    def terminal_instance(self):
        """:class:`TerminalInstance`: Terminal instance."""
        return TerminalInstance(self.__stub.GetTerminalInstance(self.msg))

    @terminal_instance.setter
    def terminal_instance(self, value):
        self.__stub.SetTerminalInstance(messages.term_inst_term_set_instance_message(self, value))



"""Primitive classes."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.layout.layout import Layout
    from ansys.edb.core.typing import ValueLike, PointLike
    from ansys.edb.core.geometry.polygon_data import PolygonData

from ansys.api.edb.v1 import board_bend_def_pb2, board_bend_def_pb2_grpc

from ansys.edb.core.inner import messages, parser
from ansys.edb.core.primitive.primitive import Primitive
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class BoardBendDef(Primitive):
    """Represents a board bend definition instance."""

    __stub: board_bend_def_pb2_grpc.BoardBendDefServiceStub = StubAccessor(StubType.board_bend_def)

    @classmethod
    def create(
        cls,
        layout: Layout,
        zone_prim: Primitive,
        bend_middle: PointLike,
        bend_radius: ValueLike,
        bend_angle: ValueLike,
    ) -> BoardBendDef:
        """Create a board bend definition.

        Parameters
        ----------
        layout : .Layout
            Layout to create the board bend definition in.
        zone_prim : .Primitive
            Zone primitive to create the board bend definition on.
        bend_middle : tuple of (:term:`Point2DLike`, :term:`Point2DLike`)
            Tuple containing the starting and ending points of the line that represents
            the middle of the bend.
        bend_radius : :term:`ValueLike`
            Radius of the bend.
        bend_angle : :term:`ValueLike`
            Angle of the bend.

        Returns
        -------
        BoardBendDef
            Board bend definition created.
        """
        return BoardBendDef(
            cls.__stub.Create(
                board_bend_def_pb2.BoardBendDefCreateMessage(
                    layout=layout.msg,
                    zone_prim=zone_prim.msg,
                    middle=messages.point_pair_message(bend_middle),
                    radius=messages.value_message(bend_radius),
                    angle=messages.value_message(bend_angle),
                )
            )
        )

    @property
    def boundary_primitive(self) -> Primitive:
        """:class:`.Primitive`: Zone primitive the board bend is placed on.

        This property is read-only.
        """
        return Primitive(self.__stub.GetBoundaryPrim(self.msg)).cast()

    @property
    @parser.to_point_data_pair
    def bend_middle(self) -> tuple[PointLike, PointLike]:
        """(:term:`Point2DLike`, :term:`Point2DLike`): Tuple of the bend middle based on starting and ending points."""
        return self.__stub.GetBendMiddle(self.msg)

    @bend_middle.setter
    def bend_middle(self, bend_middle: tuple[PointLike, PointLike]):
        self.__stub.SetBendMiddle(messages.point_pair_property_message(self, bend_middle))

    @property
    def radius(self) -> Value:
        """:class:`.Value`: Radius of the bend."""
        return Value(self.__stub.GetRadius(self.msg))

    @radius.setter
    def radius(self, val: ValueLike):
        self.__stub.SetRadius(messages.value_property_message(self, val))

    @property
    def angle(self) -> Value:
        """:class:`.Value`: Angle of the bend."""
        return Value(self.__stub.GetAngle(self.msg))

    @angle.setter
    def angle(self, val: ValueLike):
        self.__stub.SetAngle(messages.value_property_message(self, val))

    @property
    @parser.to_polygon_data_list
    def bent_regions(self) -> list[PolygonData]:
        """:obj:`list` of :class:`.PolygonData`: Bent region polygons.

        This list of a collection of polygon data represents the areas bent by the bend definition.

        This property is read-only.
        """
        return self.__stub.GetBentRegions(self.msg)



"""Stackup layer."""

from enum import Enum

import ansys.api.edb.v1.stackup_layer_pb2 as stackup_layer_pb2

from ansys.edb.core.inner import messages
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import get_stackup_layer_stub
from ansys.edb.core.utility.value import Value


class DCThicknessType(Enum):
    """Provides an enum representing DC thickness types of stackup layers."""

    EFFECTIVE = stackup_layer_pb2.HFSSSolverPropertiesMessage.EFFECTIVE
    LAYER = stackup_layer_pb2.HFSSSolverPropertiesMessage.LAYER
    MANUAL = stackup_layer_pb2.HFSSSolverPropertiesMessage.MANUAL


class RoughnessRegion(Enum):
    """Provides an enum representing regions for roughness models of stackup layers."""

    TOP = stackup_layer_pb2.LayerRoughnessRegionMessage.RoughnessRegion.TOP
    BOTTOM = stackup_layer_pb2.LayerRoughnessRegionMessage.RoughnessRegion.BOTTOM
    SIDE = stackup_layer_pb2.LayerRoughnessRegionMessage.RoughnessRegion.SIDE


class EtchNetClass(Enum):
    """Enum representing etch net classes of stackup layers."""

    NO_ETCH_POWER_GROUND = stackup_layer_pb2.EtchNetClass.NO_ETCH_POWER_GROUND
    ETCH_ALL_NETS = stackup_layer_pb2.EtchNetClass.ETCH_ALL_NETS


def _set_layer_material_name_message(layer, mat_name):
    """Convert to a ``SetLayerMaterialNameMessage`` object."""
    return stackup_layer_pb2.SetLayerMaterialMessage(layer=layer.msg, material=mat_name)


def _get_layer_material_name_message(layer, evaluated):
    """Convert to a ``GetLayerMaterialNameMessage`` object."""
    return stackup_layer_pb2.GetLayerMaterialMessage(layer=layer.msg, evaluated=evaluated)


def _stackup_layer_value_message(layer, value):
    """Convert to a ``StackupLayerValueMessage`` object."""
    return stackup_layer_pb2.StackupLayerValueMessage(
        layer=layer.msg, value=messages.value_message(value)
    )


def _layer_roughness_region_message(layer, region):
    """Convert to a ``LayerRoughnessRegionMessage`` object."""
    return stackup_layer_pb2.LayerRoughnessRegionMessage(
        layer=layer.msg, roughness_region=region.value
    )


class StackupLayer(Layer):
    """Represents a stackup layer."""

    @staticmethod
    def create(name, layer_type, thickness, elevation, material):
        """Create a stackup layer.

        Parameters
        ----------
        name : str
            Name of the stackup layer.
        layer_type : LayerType
            Type of the stackup layer.
        thickness : :term:`ValueLike`
           Thickness of the stackup layer.
        elevation : :term:`ValueLike`
            Elevation  of the stackup layer.
        material : str
            Material  of the stackup layer.

        Returns
        -------
        StackupLayer
            Stackup layer created.
        """
        params = {
            "name": name,
            "type": layer_type.value,
            "thickness": messages.value_message(thickness),
            "elevation": messages.value_message(elevation),
            "material": material,
        }

        stackup_layer = StackupLayer(
            get_stackup_layer_stub().Create(stackup_layer_pb2.StackupLayerCreationMessage(**params))
        )
        return stackup_layer

    @property
    def negative(self):
        """:obj:`bool`: Flag indicating if the layer is a negative layer."""
        return get_stackup_layer_stub().GetNegative(self.msg).value

    @negative.setter
    def negative(self, is_negative):
        get_stackup_layer_stub().SetNegative(
            stackup_layer_pb2.SetNegativeMessage(layer=self.msg, is_negative=is_negative)
        )

    @property
    def thickness(self):
        """:class:`.Value`: Thickness value of the layer.

        The setter accepts a :term:`ValueLike` term.
        """
        return Value(get_stackup_layer_stub().GetThickness(self.msg))

    @thickness.setter
    def thickness(self, thickness):
        get_stackup_layer_stub().SetThickness(_stackup_layer_value_message(self, thickness))

    @property
    def lower_elevation(self):
        """:class:`.Value`: Lower elevation value of the layer.

        The setter accepts a :term:`ValueLike` term.
        """
        return Value(get_stackup_layer_stub().GetLowerElevation(self.msg))

    @lower_elevation.setter
    def lower_elevation(self, lower_elevation):
        get_stackup_layer_stub().SetLowerElevation(
            _stackup_layer_value_message(self, lower_elevation)
        )

    @property
    def upper_elevation(self):
        """:class:`.Value`: Upper elevation value of the layer.

        This property is read-only.
        """
        return Value(get_stackup_layer_stub().GetUpperElevation(self.msg))

    def get_material(self, evaluated=True):
        """Get the name of the material of the layer.

        Parameters
        ----------
        evaluated : bool, default: True
            Whether to evaluate the material if it is parameterized.

        Returns
        -------
        str
            Material name.
        """
        return (
            get_stackup_layer_stub()
            .GetMaterial(_get_layer_material_name_message(self, evaluated))
            .value
        )

    def set_material(self, material_name):
        """Set the name of the material of the layer.

        Parameters
        ----------
        material_name : str
            New name of the material.
        """
        get_stackup_layer_stub().SetMaterial(_set_layer_material_name_message(self, material_name))

    def get_fill_material(self, evaluated=True):
        """Get the name of the fill material of the layer.

        Parameters
        ----------
        evaluated : bool, default: True
            Whether to evaluate the material if it is parameterized.

        Returns
        -------
        str
            Name of the fill material.
        """
        return (
            get_stackup_layer_stub()
            .GetFillMaterial(_get_layer_material_name_message(self, evaluated))
            .value
        )

    def set_fill_material(self, fill_material_name):
        """Set the name of the fill material of the layer.

        Parameters
        ----------
        fill_material_name : str
            New name of the fill material.
        """
        get_stackup_layer_stub().SetFillMaterial(
            _set_layer_material_name_message(self, fill_material_name)
        )

    @property
    def roughness_enabled(self):
        """:obj:`bool`: Flag indicating if roughness models are used by the layer."""
        return get_stackup_layer_stub().IsRoughnessEnabled(self.msg).value

    @roughness_enabled.setter
    def roughness_enabled(self, enable_roughness):
        get_stackup_layer_stub().SetRoughnessEnabled(
            stackup_layer_pb2.SetLayerPropEnabledMessage(layer=self.msg, enabled=enable_roughness)
        )

    def get_roughness_model(self, region):
        """Get the roughness model used by the layer.

        Parameters
        ----------
        region : RoughnessRegion

        Returns
        -------
        :term:`RoughnessModel`
        """
        request = _layer_roughness_region_message(self, region)
        response = get_stackup_layer_stub().GetRoughnessModel(request)
        roughness = Value(response.roughness)
        return (
            roughness
            if not response.HasField("huray_surface_ratio")
            else (roughness, Value(response.surface_ratio))
        )

    def set_roughness_model(self, roughness_model, region):
        """Set the roughness model used by the layer.

        Parameters
        ----------
        roughness_model : :term:`RoughnessModel`
        region : RoughnessRegion
        """
        roughness_model_msg = stackup_layer_pb2.RoughnessModelMessage()
        is_groisse_roughness = isinstance(roughness_model, Value)
        roughness_model_msg.roughness.CopyFrom(
            messages.value_message(roughness_model if is_groisse_roughness else roughness_model[0])
        )
        if not is_groisse_roughness:
            roughness_model_msg.surface_ratio.CopyFrom(messages.value_message(roughness_model[1]))
        request = stackup_layer_pb2.SetRoughnessModelMessage(
            layer_rough_region=_layer_roughness_region_message(self, region),
            roughness_model=roughness_model_msg,
        )
        get_stackup_layer_stub().SetRoughnessModel(request)

    @property
    def etch_factor_enabled(self):
        """:obj:`bool`: Flag indicating if an etch factor is used by the layer."""
        return get_stackup_layer_stub().IsEtchFactorEnabled(self.msg).value

    @etch_factor_enabled.setter
    def etch_factor_enabled(self, enable_etch_factor):
        get_stackup_layer_stub().SetEtchFactorEnabled(
            stackup_layer_pb2.SetLayerPropEnabledMessage(layer=self.msg, enabled=enable_etch_factor)
        )

    @property
    def etch_factor(self):
        """:class:`.Value`: Etch factor of the layer.

        The setter accepts a :term:`ValueLike` term.
        """
        return Value(get_stackup_layer_stub().GetEtchFactor(self.msg))

    @etch_factor.setter
    def etch_factor(self, etch_factor):
        get_stackup_layer_stub().SetEtchFactor(_stackup_layer_value_message(self, etch_factor))

    @property
    def etch_net_class(self) -> EtchNetClass:
        """:class:`.EtchNetClass`: Etch net class of the layer."""
        return EtchNetClass(get_stackup_layer_stub().GetEtchNetClass(self.msg).etch_net_class)

    @etch_net_class.setter
    def etch_net_class(self, etch_net_class: EtchNetClass):
        get_stackup_layer_stub().SetEtchNetClass(
            stackup_layer_pb2.StackupLayerEtchNetClassMessage(
                layer=self.msg, etch_net_class=etch_net_class.value
            )
        )

    @property
    def use_solver_properties(self):
        """:obj:`bool`: Flag indicating if solver properties are used by the layer."""
        return get_stackup_layer_stub().IsEtchFactorEnabled(self.msg).value

    @use_solver_properties.setter
    def use_solver_properties(self, use_solver_properties):
        get_stackup_layer_stub().SetUseSolverProperties(
            stackup_layer_pb2.SetLayerPropEnabledMessage(
                layer=self.msg, enabled=use_solver_properties
            )
        )

    @property
    def hfss_solver_properties(self):
        """:term:`HFSSSolverProperties`: HFSS solver properties of the layer."""
        response = get_stackup_layer_stub().GetHFSSSolverProperties(self.msg)
        return (
            DCThicknessType(response.dc_thickness_type),
            Value(response.dc_thickness),
            response.solve_inside,
        )

    @hfss_solver_properties.setter
    def hfss_solver_properties(self, hfss_solver_props):
        hfss_solver_props_msg = stackup_layer_pb2.HFSSSolverPropertiesMessage(
            dc_thickness_type=hfss_solver_props[0].value,
            dc_thickness=messages.value_message(hfss_solver_props[1]),
            solve_inside=hfss_solver_props[2],
        )
        request = stackup_layer_pb2.SetHFSSSolverPropertiesMessage(
            layer=self.msg, hfss_solver_props=hfss_solver_props_msg
        )

        get_stackup_layer_stub().SetHFSSSolverProperties(request)

    @property
    def referencing_via_layer_ids(self):
        r""":obj:`list`\[:obj:`int`\]: Layer IDs for all via layers referencing the layer.

        This property is read-only.
        """
        return [
            via_lyr_id
            for via_lyr_id in get_stackup_layer_stub()
            .GetReferencingViaLayerIds(self.msg)
            .ref_layer_ids
        ]



"""Dielectric material definition."""
from enum import Enum

from ansys.api.edb.v1 import dielectric_material_model_pb2_grpc

from ansys.edb.core import session
from ansys.edb.core.inner import ObjBase


class DielectricMaterialModelType(Enum):
    """Enum representing dielectric material model types."""

    DEBYE = 0
    MULTIPOLE_DEBYE = 1
    DJORDJECVIC_SARKAR = 2


class DielectricMaterialModel(ObjBase):
    """Represents a dielectric material model."""

    __stub: dielectric_material_model_pb2_grpc.DielectricMaterialModelServiceStub = (
        session.StubAccessor(session.StubType.dielectric_material_model)
    )

    @property
    def type(self) -> DielectricMaterialModelType:
        """:class:`.DielectricMaterialModelType`: Type of the dielectric material model.

        This property is read-only.
        """
        return DielectricMaterialModelType(self.__stub.GetType(self.msg).value)



"""Bondwire."""
from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.edb.core.hierarchy.cell_instance import CellInstance
    from ansys.edb.core.layout.layout import Layout
    from ansys.edb.core.typing import NetLike, ValueLike, LayerLike

from ansys.api.edb.v1 import bondwire_pb2, bondwire_pb2_grpc

from ansys.edb.core.inner import messages
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.primitive.primitive import Primitive
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class BondwireCrossSectionType(Enum):
    """Provides an enum representing bondwire cross section types."""

    ROUND = bondwire_pb2.BONDWIRE_ROUND
    RECTANGLE = bondwire_pb2.BONDWIRE_RECTANGLE
    INVALID = bondwire_pb2.INVALID_BONDWIRE_CROSS_SECTION_TYPE


class BondwireType(Enum):
    """Provides an enum representing bondwire types."""

    APD = bondwire_pb2.APD_BONDWIRE
    JEDEC4 = bondwire_pb2.JEDEC4_BONDWIRE
    JEDEC5 = bondwire_pb2.JEDEC5_BONDWIRE
    NUM_OF_TYPE = bondwire_pb2.NUM_OF_BONDWIRE_TYPE
    INVALID = bondwire_pb2.INVALID_BONDWIRE_TYPE


class Bondwire(Primitive):
    """Represents a bondwire object."""

    __stub: bondwire_pb2_grpc.BondwireServiceStub = StubAccessor(StubType.bondwire)

    @classmethod
    def create(
        cls,
        layout: Layout,
        bondwire_type: BondwireType,
        definition_name: str,
        placement_layer: str,
        width: ValueLike,
        material: str,
        start_context: CellInstance,
        start_layer_name: str,
        start_x: ValueLike,
        start_y: ValueLike,
        end_context: CellInstance,
        end_layer_name: str,
        end_x: ValueLike,
        end_y: ValueLike,
        net: NetLike | None,
    ) -> Bondwire:
        """Create a bondwire.

        Parameters
        ----------
        layout : .Layout
            Layout to create the bondwire in.
        bondwire_type : .BondwireType
            Type of the bondwire. Options are ``kAPDBondWire`` and ``kJDECBondWire``.
        definition_name : str
            Bondwire definition name.
        placement_layer : str
            Layer name to create the bondwire on.
        width : :term:`ValueLike`
            Bondwire width.
        material : str
            Bondwire material name.
        start_context : .CellInstance
            Start context :obj:`None` means top-level,.
        start_layer_name : str
            Name of the start layer.
        start_x : :term:`ValueLike`
            X value of the start point.
        start_y : :term:`ValueLike`
            Y value of the start point.
        end_context : .CellInstance
            End content :obj:`None` means top-level.
        end_layer_name : str
            Name of the end layer.
        end_x : :term:`ValueLike`
            X value of the end point.
        end_y : :term:`ValueLike`
            Y value of the end point.
        net : :term:`NetLike` or None
            Net of the bondwire.

        Returns
        -------
        Bondwire
            Bondwire object created.
        """
        return Bondwire(
            cls.__stub.Create(
                bondwire_pb2.BondwireCreateMessage(
                    layout=layout.msg,
                    net=messages.net_ref_message(net),
                    bondwire_type=bondwire_type.value,
                    definition_name=definition_name,
                    placement_layer=placement_layer,
                    width=messages.value_message(width),
                    material=material,
                    start_context=messages.edb_obj_message(start_context),
                    start_layer_name=start_layer_name,
                    start_x=messages.value_message(start_x),
                    start_y=messages.value_message(start_y),
                    end_context=messages.edb_obj_message(end_context),
                    end_layer_name=end_layer_name,
                    end_x=messages.value_message(end_x),
                    end_y=messages.value_message(end_y),
                )
            )
        )

    def get_material(self, evaluated: bool = True) -> str:
        """Get the material of the bondwire.

        Parameters
        ----------
        evaluated : bool, default: True
            Whether an evaluated (in variable namespace) material name is wanted.

        Returns
        -------
        str
            Material name.
        """
        return self.__stub.GetMaterial(Bondwire._bondwire_bool_message(self, evaluated))

    def set_material(self, material: str):
        """Set the material of the bondwire.

        Parameters
        ----------
        material : str
            Material name.
        """
        self.__stub.SetMaterial(bondwire_pb2.SetMaterialMessage(target=self.msg, material=material))

    @property
    def type(self) -> BondwireType:
        """:class:`BondwireType`: Type of the bondwire."""
        btype_msg = self.__stub.GetType(self.msg)
        return BondwireType(btype_msg.type)

    @type.setter
    def type(self, bondwire_type: BondwireType):
        self.__stub.SetType(
            bondwire_pb2.SetBondwireTypeMessage(target=self.msg, type=bondwire_type.value)
        )

    @property
    def cross_section_type(self) -> BondwireCrossSectionType:
        """:class:`BondwireCrossSectionType`: Cross-section type of the bondwire."""
        return BondwireCrossSectionType(self.__stub.GetCrossSectionType(self.msg).type)

    @cross_section_type.setter
    def cross_section_type(self, bondwire_type: BondwireCrossSectionType):
        self.__stub.SetCrossSectionType(
            bondwire_pb2.SetCrossSectionTypeMessage(target=self.msg, type=bondwire_type.value)
        )

    @property
    def cross_section_height(self) -> Value:
        """:class:`.Value`: Cross-section height of the bondwire."""
        return Value(self.__stub.GetCrossSectionHeight(self.msg))

    @cross_section_height.setter
    def cross_section_height(self, height: ValueLike):
        self.__stub.SetCrossSectionHeight(
            bondwire_pb2.SetCrossSectionHeightMessage(
                target=self.msg, height=messages.value_message(height)
            )
        )

    def get_definition_name(self, evaluated: bool = True) -> str:
        """Get the definition name of the bondwire object.

        Parameters
        ----------
        evaluated : bool, default: True
            Whether an evaluated (in variable namespace) material name is wanted.

        Returns
        -------
        str
            Bondwire definition name.
        """
        return self.__stub.GetDefinitionName(Bondwire._bondwire_bool_message(self, evaluated)).value

    def set_definition_name(self, definition_name: str):
        """Set the definition name of a bondwire.

        Parameters
        ----------
        definition_name : str
            Bondwire definition name to set.
        """
        self.__stub.SetDefinitionName(
            bondwire_pb2.SetDefinitionNameMessage(target=self.msg, definition_name=definition_name)
        )

    def get_traj(self) -> tuple[Value, Value, Value, Value]:
        """Get trajectory parameters of the bondwire.

        Returns
        -------
        tuple of (.Value, .Value, .Value, .Value)

            Returns a tuple in this format:

            **(x1, y1, x2, y2)**

            **x1** : X value of the start point.

            **y1** : Y value of the start point.

            **x1** : X value of the end point.

            **y1** : Y value of the end point.
        """
        traj_msg = self.__stub.GetTraj(self.msg)
        return (
            Value(traj_msg.x1),
            Value(traj_msg.y1),
            Value(traj_msg.x2),
            Value(traj_msg.y2),
        )

    def set_traj(self, x1: ValueLike, y1: ValueLike, x2: ValueLike, y2: ValueLike):
        """Set the parameters of the trajectory of the bondwire.

        Parameters
        ----------
        x1 : :term:`ValueLike`
            X value of the start point.
        y1 : :term:`ValueLike`
            Y value of the start point.
        x2 : :term:`ValueLike`
            X value of the end point.
        y2 : :term:`ValueLike`
            Y value of the end point.
        """
        self.__stub.SetTraj(
            bondwire_pb2.SetBondwireTrajMessage(
                target=self.msg,
                traj=bondwire_pb2.BondwireTrajMessage(
                    x1=messages.value_message(x1),
                    y1=messages.value_message(y1),
                    x2=messages.value_message(x2),
                    y2=messages.value_message(y2),
                ),
            )
        )

    @property
    def width(self) -> Value:
        """:class:`.Value`: Width of the bondwire."""
        val = self.__stub.GetWidthValue(self.msg)
        return Value(val)

    @width.setter
    def width(self, width: ValueLike):
        self.__stub.SetWidthValue(
            bondwire_pb2.BondwireValueMessage(target=self.msg, value=messages.value_message(width))
        )

    def get_start_elevation(self, start_context: CellInstance) -> Layer:
        """Get the start elevation layer of the bondwire.

        Parameters
        ----------
        start_context : .CellInstance
            Start cell context of the bondwire.

        Returns
        -------
        .Layer
            Start elevation level of the bondwire.
        """
        return Layer(
            self.__stub.GetStartElevation(Bondwire._get_elevation_message(self, start_context))
        ).cast()

    def set_start_elevation(self, start_context: CellInstance, layer: LayerLike):
        """Set the start elevation of the bondwire.

        Parameters
        ----------
        start_context : .CellInstance
            Start cell context of the bondwire. :obj:`None` means top-level.
        layer : :term:`LayerLike`
            Start layer of the bondwire.
        """
        self.__stub.SetStartElevation(Bondwire._set_elevation_message(self, start_context, layer))

    def get_end_elevation(self, end_context: CellInstance) -> Layer:
        """Get the end elevation layer of the bondwire.

        Parameters
        ----------
        end_context : .CellInstance
            End cell context of the bondwire.

        Returns
        -------
        .Layer
            End elevation layer of the bondwire.
        """
        return Layer(
            self.__stub.GetEndElevation(Bondwire._get_elevation_message(self, end_context))
        ).cast()

    def set_end_elevation(self, end_context: CellInstance, layer: LayerLike):
        """Set the end elevation of the bondwire.

        Parameters
        ----------
        end_context : .CellInstance
            End cell context of the bondwire. :obj:`None` means top-level.
        layer : :term:`LayerLike`
            End layer of the bondwire.
        """
        self.__stub.SetEndElevation(Bondwire._set_elevation_message(self, end_context, layer))

    @staticmethod
    def _bondwire_bool_message(b, evaluated):
        return bondwire_pb2.BondwireBoolMessage(target=b.msg, evaluated=evaluated)

    @staticmethod
    def _get_elevation_message(b, cell_instance):
        return bondwire_pb2.GetElevationMessage(
            bw=b.msg, cell_instance=messages.edb_obj_message(cell_instance)
        )

    @staticmethod
    def _set_elevation_message(b, cell_instance, lyrname):
        return bondwire_pb2.SetElevationMessage(
            target=Bondwire._get_elevation_message(b, cell_instance), lyrname=lyrname
        )



"""Instance array."""

from ansys.api.edb.v1.inst_array_pb2_grpc import InstArrayServiceStub

from ansys.edb.core.edb_defs import LayoutObjType
from ansys.edb.core.hierarchy import cell_instance
from ansys.edb.core.inner import messages, parser
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class InstArray(cell_instance.CellInstance):
    """Represents an instance array object."""

    __stub: InstArrayServiceStub = StubAccessor(StubType.inst_array)
    layout_obj_type = LayoutObjType.CELL_INSTANCE

    @classmethod
    def create(cls, layout, name, ref, orig, xaxis, yaxis, xcount, ycount):
        """Create an instance array with a layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to create the instance array in.
        name : str
            Name of the instance array.
        ref : :class:`.Layout`
            Layout that the instance array refers to.
        orig : :class:`.PointData`
            Point data that represents the origin of the instance array.
        xaxis : :class:`.PointData`
            Point data that represents the x axis of the instance array.
        yaxis : :class:`.PointData`
            Point data that represents the y axis of the instance array.
        xcount : :class:`.Value`
            Value for the x count of the instance array.
        ycount : :class:`.Value`
            Value for the y count of the instance array.

        Returns
        -------
        InstArray
            Instance array created.
        """
        return InstArray(
            cls.__stub.Create(
                messages.inst_array_creation_message(
                    layout, name, ref, orig, xaxis, yaxis, xcount, ycount
                )
            )
        )

    @classmethod
    def find(cls, layout, name):
        """Find an instance array by name in a given layout.

        Parameters
        ----------
        layout : :class:`.Layout`
            Layout to search for the instance array.
        name : str
            Name of the instance array.

        Returns
        -------
        InstArray
            Instance array that is found, ``None`` otherwise.
        """
        return InstArray(cls.__stub.FindByName(messages.string_property_message(layout, name)))

    @property
    @parser.to_point_data
    def orig(self):
        """:class:`.PointData`: Origin of the instance array."""
        return self.__stub.GetOrig(self.msg)

    @property
    @parser.to_point_data
    def x_axis(self):
        """:class:`.PointData`: X axis of the instance array."""
        return self.__stub.GetXAxis(self.msg)

    @x_axis.setter
    def x_axis(self, value):
        self.__stub.SetXAxis(messages.point_property_message(self, value))

    @property
    @parser.to_point_data
    def y_axis(self):
        """:class:`.PointData`: Y axis of the instance array."""
        return self.__stub.GetYAxis(self.msg)

    @y_axis.setter
    def y_axis(self, value):
        self.__stub.SetYAxis(messages.point_property_message(self, value))

    @property
    def x_count(self):
        """:class:`.Value`: X count of the instance array."""
        return Value(self.__stub.GetXCount(self.msg))

    @x_count.setter
    def x_count(self, value):
        self.__stub.SetXCount(messages.value_property_message(self, value))

    @property
    def y_count(self):
        """:class:`.Value`: Y count of the instance array."""
        return Value(self.__stub.GetYCount(self.msg))

    @y_count.setter
    def y_count(self, value):
        self.__stub.SetYCount(messages.value_property_message(self, value))

    def decompose(self):
        """Decompose the instance array."""
        self.__stub.Decompose(self.msg)



"""Hierarchy object."""

from ansys.edb.core.definition import component_def
from ansys.edb.core.inner import conn_obj, messages
from ansys.edb.core.layer.layer import Layer
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility.transform import Transform
from ansys.edb.core.utility.value import Value


class HierarchyObj(conn_obj.ConnObj):
    """Provides the base class for an hierarchy object."""

    __stub = StubAccessor(StubType.hierarchy_obj)

    @property
    def transform(self):
        """:class:`.Transform`: \
        Transformation information of the hierarchy object."""
        transform_msg = self.__stub.GetTransform(self.msg)
        return Transform(transform_msg)

    @transform.setter
    def transform(self, value):
        """Set transform."""
        self.__stub.SetTransform(messages.pointer_property_message(self, value))

    @property
    def name(self):
        """:obj:`str`: Name of the hierarchy object."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, value):
        self.__stub.SetName(messages.edb_obj_name_message(self, value))

    @property
    def component_def(self):
        """:class:`.ComponentDef`: Component definition for the \
        hierarchy object if it exists, ``None`` otherwise.

        This property is read-only.
        """
        return component_def.ComponentDef(self.__stub.GetComponent(self.msg))

    @property
    def placement_layer(self):
        """:class:`.Layer`: Placement layer for the hierarchy object."""
        return Layer(self.__stub.GetPlacementLayer(self.msg)).cast()

    @placement_layer.setter
    def placement_layer(self, value):
        self.__stub.SetPlacementLayer(messages.pointer_property_message(self, value))

    @property
    def location(self):
        """:obj:`tuple` (:class:`.Value`, \
        :class:`.Value`): \
        Location [x, y] of the hierarchy object on the :obj:`placement_layer` object."""
        pnt_msg = self.__stub.GetLocation(self.msg)
        return [Value(pnt_msg.x), Value(pnt_msg.y)]

    @location.setter
    def location(self, value):
        self.__stub.SetLocation(messages.point_property_message(self, value))

    @property
    def solve_independent_preference(self):
        """:obj:`bool`: Flag indicating if the object is assigned to solve independent of its parent context.

        Returns
        -------
        bool
            ``True`` if the object is assigned to solve independently, ``False`` if the object is embedded.

        Notes
        -----
        For a :class:`.ComponentModel`
        instance, this flag indicates if the model is embedded with the field solver, when applicable.

        For a :class:`.CellInstance` instance,
        it indicates if the design's geometry is flattened/meshed with the parent, when applicable.
        """
        return self.__stub.GetSolveIndependentPreference(self.msg).value

    @solve_independent_preference.setter
    def solve_independent_preference(self, value):
        self.__stub.SetSolveIndependentPreference(messages.bool_property_message(self, value))



"""Mesh operation."""


class MeshOperation:
    """Represents mesh operations.

    Attributes
    ----------
    name : str, default: ""
        Name of the mesh operation. The default is ``""``, in which case a name is automatically assigned.
    net_layer_info : list[tuple(str, str, bool)], default: None
        List with each entry having a net name, layer name, and ``isSheet`` flag indicating
        if the entry is a sheet.
    enabled : bool, default: True
        Whether the mesh operation is enabled.
    refine_inside : bool, default: False
        Whether to refine inside.
    mesh_region : str, default: ""
        Mesh region.
    solve_inside : bool, default: False
        Whether to solve inside.
    """

    def __init__(
        self,
        name="",
        net_layer_info=None,
        enabled=True,
        refine_inside=False,
        mesh_region="",
        solve_inside=False,
    ):
        """Create a mesh operation."""
        self._name = name
        self._net_layer_info = [] if net_layer_info is None else net_layer_info
        self._enabled = enabled
        self._refine_inside = refine_inside
        self._mesh_region = mesh_region
        self._solve_inside = solve_inside

    @property
    def name(self):
        """:obj:`str`: Name of the mesh operation."""
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def net_layer_info(self):
        """:obj:`list` of :obj:`tuple` of (:obj:`str`, :obj:`str`, :obj:`bool`): List of net layer \
        information for the mesh operation.

        The tuple is in this form: (net_name, layer_name, is_sheet)``.
        """
        return self._net_layer_info

    @net_layer_info.setter
    def net_layer_info(self, net_layer_info):
        self._net_layer_info = net_layer_info

    @property
    def enabled(self):
        """:obj:`bool`: Flag indicating if the mesh operation is enabled."""
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        self._enabled = enabled

    @property
    def refine_inside(self):
        """:obj:`bool`: Flag indicating if refining inside is enabled."""
        return self._refine_inside

    @refine_inside.setter
    def refine_inside(self, refine_inside):
        self._refine_inside = refine_inside

    @property
    def mesh_region(self):
        """:obj:`str`: Name of the mesh region."""
        return self._mesh_region

    @mesh_region.setter
    def mesh_region(self, mesh_region):
        self._mesh_region = mesh_region

    @property
    def solve_inside(self):
        """:obj:`bool`: Flag indicating if solve inside is enabled."""
        return self._solve_inside

    @solve_inside.setter
    def solve_inside(self, solve_inside):
        self._solve_inside = solve_inside


class SkinDepthMeshOperation(MeshOperation):
    """Represents a skin depth mesh operation.

    Attributes
    ----------
    name : str, default: ""
        Name of the mesh operation. The default is ``""``, in which case a name is automatically assigned.
    net_layer_info : list[tuple(str, str, bool)], default: None
        List with each entry having a net name, layer name, and ``isSheet`` flag indicating
        if the entry is a sheet.
    enabled : bool, default: True
        Whether the mesh operation is enabled.
    refine_inside : bool, default: False
        Whether to refine inside.
    mesh_region : str, default: ""
        Mesh region.
    solve_inside : bool, default: False
        Whether to solve inside.
    skin_depth : str, default: "1um"
        Skin depth.`.
    surface_triangle_length : str, default: "1mm"
        Surface triangle length with units.
    num_layers : str, default: 2
        Number of layers.
    max_elements : str, default: "1000"
        Maximum number of elements.
    restrict_max_elements : bool, default: False
        Whether to restrict the number of elements.
    """

    def __init__(
        self,
        name="",
        net_layer_info=None,
        enabled=True,
        refine_inside=False,
        mesh_region="",
        solve_inside=False,
        skin_depth="1um",
        surface_triangle_length="1mm",
        num_layers="2",
        max_elements="1000",
        restrict_max_elements=False,
    ):
        """Create a skin depth mesh operation."""
        super().__init__(name, net_layer_info, enabled, refine_inside, mesh_region, solve_inside)
        self._skin_depth = skin_depth
        self._surface_triangle_length = surface_triangle_length
        self._num_layers = num_layers
        self._max_elements = max_elements
        self._restrict_max_elements = restrict_max_elements

    @property
    def skin_depth(self):
        """:obj:`str`: Skin depth."""
        return self._skin_depth

    @skin_depth.setter
    def skin_depth(self, skin_depth):
        self._skin_depth = skin_depth

    @property
    def surface_triangle_length(self):
        """:obj:`str`: Surface triangle length."""
        return self._surface_triangle_length

    @surface_triangle_length.setter
    def surface_triangle_length(self, surface_triangle_length):
        self._surface_triangle_length = surface_triangle_length

    @property
    def number_of_layers(self):
        """:obj:`str`: Number of layers."""
        return self._num_layers

    @number_of_layers.setter
    def number_of_layers(self, number_of_layers):
        self._num_layers = number_of_layers

    @property
    def max_elements(self):
        """:obj:`str`: Maximum number of mesh elements."""
        return self._max_elements

    @max_elements.setter
    def max_elements(self, max_elements):
        self._max_elements = max_elements

    @property
    def restrict_max_elements(self):
        """:obj:`bool`: Flag indicating if the maximum number of mesh elements is restricted."""
        return self._restrict_max_elements

    @restrict_max_elements.setter
    def restrict_max_elements(self, restrict_max_elements):
        self._restrict_max_elements = restrict_max_elements


class LengthMeshOperation(MeshOperation):
    """Represents a length mesh operation.

    Attributes
    ----------
    name : str, default: ""
        Name of the mesh operation. The default is ``""``, in which case a name is automatically assigned.
    net_layer_info : list[tuple(str, str, bool)], default: None
        List with each entry having a net name, layer name, and ``isSheet`` flag indicating
        if the entry is a sheet.
    enabled : bool, default: True
        Whether the mesh operation is enabled.
    refine_inside : bool, default: False
        Whether to refine inside.
    mesh_region : str, default: ""
        Mesh region.
    solve_inside : bool, default: False
        Whether to solve inside.
    max_length : str, default: "1mm"
        maximum length of the mesh elements.
    restrict_max_length : str, default: True
        Whether to restrict the length of the mesh elements.
    max_elements : str, default: "1000"
        Number of layers.
    restrict_max_elements : bool, default: False
        Whether to restrict the number of elements.
    """

    def __init__(
        self,
        name="",
        net_layer_info=None,
        enabled=True,
        refine_inside=False,
        mesh_region="",
        solve_inside=False,
        max_length="1mm",
        restrict_max_length=True,
        max_elements="1000",
        restrict_max_elements=False,
    ):
        """Initialize an instance of a skin depth mesh operation."""
        super().__init__(name, net_layer_info, enabled, refine_inside, mesh_region, solve_inside)
        self._max_length = max_length
        self._restrict_max_length = restrict_max_length
        self._max_elements = max_elements
        self._restrict_max_elements = restrict_max_elements

    @property
    def max_length(self):
        """:obj:`str`: Maximum length of the mesh elements."""
        return self._max_length

    @max_length.setter
    def max_length(self, max_length):
        self._max_length = max_length

    @property
    def restrict_max_length(self):
        """:obj:`bool`: Flag indicating if the maximum length of mesh elements is restricted."""
        return self._restrict_max_length

    @restrict_max_length.setter
    def restrict_max_length(self, restrict_max_length):
        self._restrict_max_length = restrict_max_length

    @property
    def max_elements(self):
        """:obj:`str`: Maximum number of mesh elements."""
        return self._max_elements

    @max_elements.setter
    def max_elements(self, max_elements):
        self._max_elements = max_elements

    @property
    def restrict_max_elements(self):
        """:obj:`bool`: Flag indicating if the maximum number of mesh elements is restricted."""
        return self._restrict_max_elements

    @restrict_max_elements.setter
    def restrict_max_elements(self, restrict_max_elements):
        self._restrict_max_elements = restrict_max_elements



"""EDB-Core Python package for the Electronics Database (EDB) format."""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version("ansys-edb-core")



"""Padstack definition data."""

from __future__ import annotations

from enum import Enum

import ansys.api.edb.v1.padstack_def_data_pb2 as pb
from ansys.api.edb.v1.padstack_def_data_pb2_grpc import PadstackDefDataServiceStub
import google.protobuf.empty_pb2 as empty_pb2

from ansys.edb.core.inner import ObjBase, messages, parser
from ansys.edb.core.session import StubAccessor, StubType
from ansys.edb.core.utility import conversions
from ansys.edb.core.utility.value import Value


class PadType(Enum):
    """Provides an enum representing pad types."""

    REGULAR_PAD = pb.REGULAR_PAD
    ANTI_PAD = pb.ANTI_PAD
    THERMAL_PAD = pb.THERMAL_PAD
    HOLE = pb.HOLE
    UNKNOWN_GEOM_TYPE = pb.UNKNOWN_GEOM_TYPE


class PadGeometryType(Enum):
    """Provides an enum representing pad geometry types."""

    PADGEOMTYPE_NO_GEOMETRY = pb.PADGEOMTYPE_NO_GEOMETRY
    PADGEOMTYPE_CIRCLE = pb.PADGEOMTYPE_CIRCLE
    PADGEOMTYPE_SQUARE = pb.PADGEOMTYPE_SQUARE
    PADGEOMTYPE_RECTANGLE = pb.PADGEOMTYPE_RECTANGLE
    PADGEOMTYPE_OVAL = pb.PADGEOMTYPE_OVAL
    PADGEOMTYPE_BULLET = pb.PADGEOMTYPE_BULLET
    PADGEOMTYPE_NSIDED_POLYGON = pb.PADGEOMTYPE_NSIDED_POLYGON
    PADGEOMTYPE_POLYGON = pb.PADGEOMTYPE_POLYGON
    PADGEOMTYPE_ROUND45 = pb.PADGEOMTYPE_ROUND45
    PADGEOMTYPE_ROUND90 = pb.PADGEOMTYPE_ROUND90
    PADGEOMTYPE_SQUARE45 = pb.PADGEOMTYPE_SQUARE45
    PADGEOMTYPE_SQUARE90 = pb.PADGEOMTYPE_SQUARE90
    PADGEOMTYPE_INVALID_GEOMETRY = pb.PADGEOMTYPE_INVALID_GEOMETRY


class PadstackHoleRange(Enum):
    """Provides an enum representing pad hole ranges."""

    THROUGH = pb.THROUGH
    BEGIN_ON_UPPER_PAD = pb.BEGIN_ON_UPPER_PAD
    END_ON_LOWER_PAD = pb.END_ON_LOWER_PAD
    UPPER_PAD_TO_LOWER_PAD = pb.UPPER_PAD_TO_LOWER_PAD
    UNKNOWN_RANGE = pb.UNKNOWN_RANGE


class SolderballShape(Enum):
    """Provides an enum representing solder ball shapes."""

    NO_SOLDERBALL = pb.NO_SOLDERBALL
    SOLDERBALL_CYLINDER = pb.SOLDERBALL_CYLINDER
    SOLDERBALL_SPHEROID = pb.SOLDERBALL_SPHEROID
    UNKNOWN_SOLDERBALL_SHAPE = pb.UNKNOWN_SOLDERBALL_SHAPE


class SolderballPlacement(Enum):
    """Provides an enum representing solder ball placement."""

    ABOVE_PADSTACK = pb.ABOVE_PADSTACK
    BELOW_PADSTACK = pb.BELOW_PADSTACK
    UNKNOWN_PLACEMENT = pb.UNKNOWN_PLACEMENT


class ConnectionPtDirection(Enum):
    """Provides an enum representing connection pt direction."""

    PS_NO_DIRECTION = pb.PS_NO_DIRECTION
    PS_ANY_DIRECTION = pb.PS_ANY_DIRECTION
    PS_0_DIRECTION = pb.PS_0_DIRECTION
    PS_45_DIRECTION = pb.PS_45_DIRECTION
    PS_90_DIRECTION = pb.PS_90_DIRECTION
    PS_135_DIRECTION = pb.PS_135_DIRECTION
    PS_180_DIRECTION = pb.PS_180_DIRECTION
    PS_225_DIRECTION = pb.PS_225_DIRECTION
    PS_270_DIRECTION = pb.PS_270_DIRECTION
    PS_315_DIRECTION = pb.PS_315_DIRECTION
    PS_UNKNOWN_DIRECTION = pb.PS_UNKNOWN_DIRECTION


class PadstackDefData(ObjBase):
    """Represents a padstack data definition."""

    __stub: PadstackDefDataServiceStub = StubAccessor(StubType.padstack_def_data)

    @classmethod
    def create(cls):
        """
        Create a padstack data definition.

        Returns
        -------
        PadstackDefData
            Padstack data definition created.
        """
        return PadstackDefData(cls.__stub.Create(empty_pb2.Empty()))

    @property
    def material(self):
        """:obj:`str`: Material name of the hole of the padstack definition data object."""
        return self.__stub.GetMaterial(self.msg)

    @material.setter
    def material(self, name):
        self.__stub.SetMaterial(pb.PadstackDefDataSetMaterialMessage(target=self.msg, name=name))

    @property
    def layer_names(self):
        """:obj:`list` of :obj:`str`: List of layer names in the padstack definition data object.

        This property is read-only.
        """
        layer_names_msg = self.__stub.GetLayerNames(self.msg).names
        return list(layer_names_msg)

    @property
    def layer_ids(self):
        """:obj:`list` of :obj:`int`: All layer IDs in the padstack definition data object.

        This property is read-only.
        """
        layer_ids_msg = self.__stub.GetLayerIds(self.msg)
        return list(layer_ids_msg.ids)

    def add_layers(self, names):
        """
        Add layers to the padstack definition data object.

        Parameters
        ----------
        names : List[str]
            List of layer names.
        """
        return self.__stub.AddLayers(
            pb.PadstackDefDataAddLayersMessage(
                target=self.msg, layer_names=pb.PadstackDefDataGetLayerNamesMessage(names=names)
            )
        )

    def get_pad_parameters(self, layer, pad_type):
        """
        Get pad parameters by layer name and pad type in their original values in the database.

        Parameters
        ----------
        layer : Union[str, int, None]
            Layer name.
        pad_type : PadType
            Pad type.

        Returns
        -------
        tuple[:class:`PadGeometryType`, list of :class:`.Value`, \
        :class:`.Value`, :class:`.Value`,
        :class:`.Value`]

        or

        tuple[:class:`.PolygonData`, \
        :class:`.Value`, \
        :class:`.Value`, :class:`.Value`]

            The tuple is in this format for other than polygons: ``(pad_type, sizes, offset_x, offset_y, rotation)``.

            For polygons, the tuple is in this format: ``(fp, offset_x, offset_y, rotation)``.

            - ``pad_type``: Pad type
            - ``sizes``: Pad parameters
            - ``offset_x``: X offset
            - ``offset_y``: Y offset
            - ``rotation``: Rotation
            - ``fp``: Polygon geometry
        """
        message = self.__stub.GetPadParameters(
            PadstackDefData._padstack_def_data_get_pad_parameters_message(self, layer, pad_type)
        )
        if message.HasField("generic"):
            return (
                PadGeometryType(message.generic.geometry_type),
                [Value(s) for s in message.generic.sizes],
                Value(message.generic.offset_x),
                Value(message.generic.offset_y),
                Value(message.generic.rotation),
            )
        elif message.HasField("polygon"):
            return (
                parser.to_polygon_data(message.polygon.fp),
                Value(message.polygon.offset_x),
                Value(message.polygon.offset_y),
                Value(message.polygon.rotation),
            )
        else:
            return ()

    def set_pad_parameters(
        self, layer, pad_type, offset_x, offset_y, rotation, type_geom=None, sizes=None, fp=None
    ):
        """
        Set pad parameters by layer name and pad type in their original values in the database.

        Parameters
        ----------
        layer : Union[str, int, None]
            Layer name.
        pad_type : PadType
            Pad type.
        offset_x : :class:`.Value`
            X offset.
        offset_y : :class:`.Value`
            Y offset.
        rotation : :class:`.Value`
            Rotation.
        type_geom : PadGeometryType, default: None
            Pad geometry type. The default is ``None`` if setting polygonal pad parameters.
        sizes : List[:class:`.Value`], default: None
            List of pad sizes. The default is ``None`` if setting polygonal pad parameters.
        fp : :class:`.PolygonData`, default: None
            Polygon geometry. The default is ``None`` if not setting polygonal pad parameters.
        """
        p1 = PadstackDefData._padstack_def_data_get_pad_parameters_message(self, layer, pad_type)
        message = None
        if fp is None:
            p2 = pb.PadstackDefDataGetPadParametersParametersMessage(
                geometry_type=type_geom.value,
                sizes=[messages.value_message(val) for val in sizes],
                offset_x=messages.value_message(offset_x),
                offset_y=messages.value_message(offset_y),
                rotation=messages.value_message(rotation),
            )
            message = pb.PadstackDefDataPadParametersSetMessage(
                generic=pb.PadstackDefDataSetPadParametersMessage(
                    params1=p1,
                    params2=p2,
                )
            )
        else:
            message = pb.PadstackDefDataPadParametersSetMessage(
                polygon=pb.PadstackDefDataSetPolygonalPadParametersMessage(
                    params1=p1,
                    fp=messages.polygon_data_message(fp),
                    offset_x=messages.value_message(offset_x),
                    offset_y=messages.value_message(offset_y),
                    rotation=messages.value_message(rotation),
                )
            )

        self.__stub.SetPadParameters(message)

    def get_hole_parameters(self):
        """
        Get hole parameters in their original values in the database.

        Returns
        -------
        tuple[:class:`.PolygonData`, \
        :class:`.Value`, \
        :class:`.Value`, :class:`.Value`]
            The tuple is in this format: ``(fp, offset_x, offset_y, rotation)``.

            - ``fp``: Polygon geometry
            - ``offset_x``: X offset
            - ``offset_y``: Y offset
            - ``rotation`` : Rotation
        """
        return self.get_pad_parameters(None, PadType.HOLE)

    def set_hole_parameters(
        self, offset_x, offset_y, rotation, type_geom=None, sizes=None, fp=None
    ):
        """
        Set hole parameters.

        Parameters
        ----------
        offset_x : :class:`.Value`
            X offset.
        offset_y : :class:`.Value`
            Y offset.
        rotation : :class:`.Value`
            Rotation.
        type_geom : PadGeometryType, default: None
            Pad geometry type. The default is ``None`` if setting polygonal pad parameters.
        sizes : List[:class:`.Value`], default: None
            List of pad sizes. The default is ``None`` if setting polygonal pad parameters.
        fp : :class:`.PolygonData`, default: None
            Polygon geometry. The default is ``None`` if not setting polygonal pad parameters.
        """
        return self.set_pad_parameters(
            -1, PadType.HOLE, offset_x, offset_y, rotation, type_geom, sizes, fp
        )

    @property
    def hole_range(self):
        """:class:`PadstackHoleRange`: Hole range of the padstack data definition."""
        return PadstackHoleRange(self.__stub.GetHoleRange(self.msg).hole_range)

    @hole_range.setter
    def hole_range(self, hole_range):
        self.__stub.SetHoleRange(
            pb.PadstackDefDataSetHoleRangeMessage(target=self.msg, hole_range=hole_range.value)
        )

    @property
    def plating_percentage(self):
        """:class:`.Value`: Hole plating percentage."""
        return Value(self.__stub.GetPlatingPercentage(self.msg))

    @plating_percentage.setter
    def plating_percentage(self, plating_percentage):
        self.__stub.SetPlatingPercentage(
            pb.PadstackDefDataSetPlatingPercentage(
                target=self.msg, plating_percentage=messages.value_message(plating_percentage)
            )
        )

    @property
    def solder_ball_shape(self):
        """:class:`SolderballShape`: Solder ball shape."""
        return SolderballShape(self.__stub.GetSolderBallShape(self.msg).solderball_shape)

    @solder_ball_shape.setter
    def solder_ball_shape(self, solderball_shape):
        self.__stub.SetSolderBallShape(
            pb.PadstackDefDataSetSolderballShapeMessage(
                target=self.msg, solderball_shape=solderball_shape.value
            )
        )

    @property
    def solder_ball_placement(self):
        """:class:`SolderballPlacement`: Solder ball placement or orientation."""
        return SolderballPlacement(self.__stub.GetSolderBallPlacement(self.msg))

    @solder_ball_placement.setter
    def solder_ball_placement(self, solderball_placement):
        self.__stub.SetSolderBallPlacement(
            pb.PadstackDefDataSetSolderballPlacementMessage(
                target=self.msg, solderball_placement=solderball_placement.value
            )
        )

    @property
    def solder_ball_param(self):
        """:obj:`tuple` of [:class:`.Value`, \
        :class:`.Value`]: Solder ball parameters ``(d1, d2)`` \
        in their original values in the database.

        - ``d1`` is the diameter for a cylinder solder ball or the top diameter for a spheroid
          solder ball.
        - ``d2`` is the middle diameter for a spheroid solder ball. It is not used for a cylinder solder ball.
        """
        params = self.__stub.GetSolderBallParam(self.msg)
        return (
            Value(params.d1),
            Value(params.d2),
        )

    @solder_ball_param.setter
    def solder_ball_param(self, params):
        self.__stub.SetSolderBallParam(
            pb.PadstackDefDataSetSolderBallParamMessage(
                target=self.msg,
                d1=messages.value_message(params[0]),
                d2=messages.value_message(params[1]),
            )
        )

    @property
    def solder_ball_material(self):
        """:obj:`str`: Name of the solder ball material."""
        return self.__stub.GetSolderBallMaterial(self.msg)

    @solder_ball_material.setter
    def solder_ball_material(self, material):
        self.__stub.SetSolderBallMaterial(
            pb.PadstackDefDataSetSolderBallMaterialMessage(target=self.msg, material=material)
        )

    def get_connection_pt(self, layer):
        """
        Get connection point position and direction by layer name.

        Parameters
        ----------
        layer : str
            Layer name.

        Returns
        -------
        tuple[:class:`.PointData`, :class:`ConnectionPtDirection`]
            The tuple is in a ``(position, direction)`` format:

            - ``position``: Position of the connection point.
            - ``direction``: Direction of the connection point.
        """
        msg = self.__stub.GetConnectionPt(
            pb.PadstackDefDataGetConnectionPtMessage(target=self.msg, layer=layer)
        )
        return parser.to_point_data(msg), ConnectionPtDirection(msg.direction)

    def set_connection_pt(self, layer, position, direction):
        """
        Set connection point position and direction.

        Parameters
        ----------
        layer : str
            Layer name.
        position : ansys.edb.core.typing.PointLike
            Position.
        direction : :class:`ConnectionPtDirection`
            Direction.
        """
        pos = conversions.to_point(position)
        self.__stub.SetConnectionPt(
            pb.PadstackDefDataSetConnectionPtMessage(
                target=self.msg,
                layer=layer,
                x=messages.value_message(pos.x),
                y=messages.value_message(pos.y),
                direction=direction.value,
            )
        )

    @staticmethod
    def _padstack_def_data_get_pad_parameters_message(target, layer, pad_type):
        return pb.PadstackDefDataGetPadParametersMessage(
            target=target.msg,
            layer_name=layer if isinstance(layer, str) else None,
            layer_id=layer if isinstance(layer, int) else None,
            pad_type=pad_type.value,
        )



"""HFSSPI simulation setup."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ansys.edb.core.simulation_setup.hfss_pi_simulation_settings import HFSSPISimulationSettings
from ansys.edb.core.simulation_setup.simulation_setup import SimulationSetup, SimulationSetupType

if TYPE_CHECKING:
    from ansys.edb.core.layout.cell import Cell


class HFSSPISimulationSetup(SimulationSetup):
    """Represents HFSSPI simulation setup data."""

    @classmethod
    def create(cls, cell: Cell, name: str) -> HFSSPISimulationSetup:
        """Create a HFSSPI simulation setup.

        Parameters
        ----------
        cell : :class:`.Cell`
            Cell to create the simulation setup in.
        name : str
            Name of the simulation setup.

        Returns
        -------
        HFSSPISimulationSetup
            HFSSPI simulation setup created.
        """
        return super()._create(cell, name, SimulationSetupType.HFSS_PI)

    @property
    def settings(self) -> HFSSPISimulationSettings:
        """:class:`.HFSSPISimulationSettings`: Simulation settings of the HFSSPI simulation setup."""
        return HFSSPISimulationSettings(self)



"""Layout object instance 2D geometry."""

from ansys.api.edb.v1.layout_obj_instance_2d_geometry_pb2 import GetPolygonDataMessage

from ansys.edb.core.inner.parser import to_polygon_data
from ansys.edb.core.layout_instance.layout_obj_instance_geometry import LayoutObjInstanceGeometry
from ansys.edb.core.session import LayoutObjInstance2DGeometryServiceStub, StubAccessor, StubType


class LayoutObjInstance2DGeometry(LayoutObjInstanceGeometry):
    """Represents a layout object instance 2D geometry."""

    __stub: LayoutObjInstance2DGeometryServiceStub = StubAccessor(
        StubType.layout_obj_instance_2d_geometry
    )

    @property
    def is_negative(self):
        """:obj:`bool`: Flag indicating if the geometry is negative.

        This property is read-only.
        """
        return self.__stub.IsNegative(self.msg).value

    @to_polygon_data
    def get_polygon_data(self, apply_negatives=False):
        """Get the underlying polygon data of the geometry.

        Returns
        -------
        :class:`.PolygonData`
        """
        return self.__stub.GetPolygonData(
            GetPolygonDataMessage(layout_obj_inst_geom=self.msg, apply_neg=apply_negatives)
        )



"""Component pin Definition."""
from ansys.api.edb.v1.component_pin_pb2_grpc import ComponentPinServiceStub

from ansys.edb.core.definition import component_def
from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import StubAccessor, StubType


class ComponentPin(ObjBase):
    """Represents a component pin."""

    __stub: ComponentPinServiceStub = StubAccessor(StubType.component_pin)

    @classmethod
    def create(cls, comp_def, name):
        """Create a component pin in a given component definition.

        Parameters
        ----------
        comp_def : :class:`.ComponentDef`
            Component definition to create the component pin in.
        name : str
            Name of the component pin.

        Returns
        -------
        ComponentPin
            Component pin created.
        """
        return ComponentPin(cls.__stub.Create(messages.edb_obj_name_message(comp_def, name)))

    @classmethod
    def find(cls, comp_def, name):
        """Find a component pin in a given component definition.

        Parameters
        ----------
        comp_def : :class:`.ComponentDef`
            Component definition to search for the component pin.
        name : str
            Name of the component pin.

        Returns
        -------
        ComponentPin
            Component pin found, ``None`` otherwise.
        """
        return ComponentPin(cls.__stub.FindByName(messages.edb_obj_name_message(comp_def, name)))

    @property
    def name(self):
        """:obj:`str`: Name of the component pin."""
        return self.__stub.GetName(self.msg).value

    @name.setter
    def name(self, value):
        self.__stub.SetName(messages.string_property_message(self, value))

    @property
    def number(self):
        """:obj:`int`: Serial number of the component pin inside its component definition.

        This property is read-only.
        """
        return self.__stub.GetNumber(self.msg).value

    @property
    def component_def(self):
        """:class:`.ComponentDef`: Component definition that the component pin belongs to.

        This property is read-only.
        """
        return component_def.ComponentDef(self.__stub.GetComponentDef(self.msg))



"""Via layer."""

import ansys.api.edb.v1.via_layer_pb2 as via_layer_pb2

from ansys.edb.core.layer.stackup_layer import StackupLayer
from ansys.edb.core.session import get_via_layer_stub


def _via_lyr_ref_lyr_id_msg(lyr, is_upper_ref):
    """Convert to a ``ViaLayerRefLayerIdMessage`` object."""
    return via_layer_pb2.ViaLayerRefLayerIdMessage(via_layer=lyr.msg, is_upper_ref=is_upper_ref)


class ViaLayer(StackupLayer):
    """Represents a via layer."""

    @staticmethod
    def create(name, lr_layer, ur_layer, material):
        """Create a via layer.

        Parameters
        ----------
        name : str
            Name of the via layer.
        lr_layer : str
        ur_layer : str
        material : str

        Returns
        -------
        ViaLayer
            Via layer created.
        """
        params = {
            "via_layer_name": name,
            "lower_ref_layer_name": lr_layer,
            "upper_ref_layer_name": ur_layer,
            "material_name": material,
        }
        via_layer = ViaLayer(
            get_via_layer_stub().Create(via_layer_pb2.ViaLayerCreationMessage(**params))
        )
        return via_layer

    def get_ref_layer_name(self, upper_ref):
        """Get the name of the reference layer of the via layer.

        Parameters
        ----------
        upper_ref : bool
            Whether to get the name of the upper or lower reference layer.

        Returns
        -------
        str
            Name of the reference layer.
        """
        return get_via_layer_stub().GetRefLayerName(_via_lyr_ref_lyr_id_msg(self, upper_ref)).value

    def set_ref_layer(self, ref_layer, upper_ref):
        """Set the reference layer of the via layer.

        Parameters
        ----------
        ref_layer : StackupLayer
            Layer to set as the new reference layer of the via layer.
        upper_ref : bool
            Whether to set the new reference layer as the
            upper or lower reference layer.
        """
        get_via_layer_stub().SetRefLayer(
            via_layer_pb2.SetViaLayerRefLayerMessage(
                via_layer_ref_layer_id=_via_lyr_ref_lyr_id_msg(self, upper_ref),
                new_ref_layer=ref_layer.msg,
            )
        )



"""Material definition."""
from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Tuple

if TYPE_CHECKING:
    from ansys.edb.core.typing import ValueLike
    from ansys.edb.core.database import Database

from enum import Enum

import ansys.api.edb.v1.material_def_pb2 as pb

from ansys.edb.core.definition.dielectric_material_model import DielectricMaterialModel
from ansys.edb.core.definition.material_property_thermal_modifier import (
    MaterialPropertyThermalModifier,
)
from ansys.edb.core.edb_defs import DefinitionObjType
from ansys.edb.core.inner import ObjBase, messages
from ansys.edb.core.session import MaterialDefServiceStub, StubAccessor, StubType
from ansys.edb.core.utility.value import Value


class MaterialProperty(Enum):
    """Enum representing material property types."""

    PERMITTIVITY = pb.PERMITTIVITY
    PERMEABILITY = pb.PERMEABILITY
    CONDUCTIVITY = pb.CONDUCTIVITY
    DIELECTRIC_LOSS_TANGENT = pb.DIELECTRIC_LOSS_TANGENT
    MAGNETIC_LOSS_TANGENT = pb.MAGNETIC_LOSS_TANGENT
    THERMAL_CONDUCTIVITY = pb.THERMAL_CONDUCTIVITY
    MASS_DENSITY = pb.MASS_DENSITY
    SPECIFIC_HEAT = pb.SPECIFIC_HEAT
    YOUNGS_MODULUS = pb.YOUNGS_MODULUS
    POISSONS_RATIO = pb.POISSONS_RATIO
    THERMAL_EXPANSION_COEFFICIENT = pb.THERMAL_EXPANSION_COEFFICIENT
    INVALID_PROPERTY = pb.INVALID_PROPERTY


class MaterialDef(ObjBase):
    """Represents a material and all its properties."""

    __stub: MaterialDefServiceStub = StubAccessor(StubType.material)

    @classmethod
    def create(cls, database: Database, name: str, **kwargs: Dict[str, ValueLike]) -> MaterialDef:
        """Create a material definition in a given database.

        Parameters
        ----------
        database : .Database
            Database to create the material definition in.
        name : str
            Name of the material definition.
        kwargs : dict of { str : :term:`ValueLike` }
            Dictionary of material property values.
            The dictionary key is the material property name. The dictionary value is the \
            material property value. The expected keys for the kwargs are:

            - ``"permittivity"``
            - ``"permeability"``
            - ``"conductivity"``
            - ``"dielectric_loss_tangent"``
            - ``"magnetic_loss_tangent"``
            - ``"thermal_conductivity"``
            - ``"mass_density"``
            - ``"specific_heat"``
            - ``"youngs_modulus"``
            - ``"poissons_ratio"``
            - ``"thermal_expansion_coefficient"``

        Returns
        -------
        .MaterialDef
        """
        return MaterialDef(
            cls.__stub.Create(
                pb.MaterialDefCreationMessage(
                    database=messages.edb_obj_message(database),
                    name=name,
                    properties=messages.material_properties_message(**kwargs),
                )
            )
        )

    @classmethod
    def find_by_name(cls, database: Database, name: str) -> MaterialDef:
        """Find a material definition by name in a given database.

        Parameters
        ----------
        database : .Database
            Database to search for the material definition.
        name : str
            Name of the material definition.

        Returns
        -------
        .MaterialDef
            Material definition object found.
            If a material definition isn't found, the returned material definition is :meth:`null <.is_null>`.
        """
        return MaterialDef(cls.__stub.FindByName(messages.edb_obj_name_message(database, name)))

    @property
    def definition_type(self) -> DefinitionObjType:
        """:class:`.DefinitionObjType`: Definition object type.

        This property is read-only.
        """
        return DefinitionObjType.MATERIAL_DEF

    @property
    def name(self):
        """:obj:`str`: Name of the material definition.

        This property is read-only.
        """
        return self.__stub.GetName(messages.edb_obj_message(self)).value

    @property
    def dielectric_material_model(self):
        """:class:`.DielectricMaterialModel`: \
        Dielectric material model of the material definition."""
        return DielectricMaterialModel(
            self.__stub.GetDielectricMaterialModel(messages.edb_obj_message(self))
        )

    @dielectric_material_model.setter
    def dielectric_material_model(self, dielectric):
        self.__stub.SetDielectricMaterialModel(messages.pointer_property_message(self, dielectric))

    def delete(self):
        """Delete the material definition."""
        self.__stub.Delete(messages.edb_obj_message(self))

    def set_property(
        self,
        material_property: MaterialProperty,
        value: ValueLike,
        component_id: int = None,
        col: int = None,
        row: int = None,
    ):
        """Set a material property value.

            Material properties can be defined in 3 ways:
                1. Simple: A constant value.
                    - To define a simple material property, the ``material_property`` \
                    and ``value`` parameters should be provided.
                2. Anisotropic: A 3x3 tensor consisting of only diagonal entries.
                    - To define a component of an anisotropic material property, the ``material_property``, \
                    ``value``, and ``component_id`` parameters should be provided.
                    - ``component_id`` specifies the \
                    :term:`anisotropic component ID <Anisotropic Material Property Component IDs>` \
                    of the component to set the value of.
                3. Tensor: A 3x3 tensor consisting of diagonal and off-diagonal entries.
                    - To define an entry in a tensor material property, the ``material_property``, \
                    ``value``, ``col`` and ``row`` parameters should be provided. The entry at ``T[row,col]`` \
                    will be set to the provided ``value``.

        Parameters
        ----------
        material_property : .MaterialProperty
            ID of the material property.
        value : :term:`ValueLike`
            New value.
        component_id : int, default: None
            ID of the anisotropic component (only used for anisotropic material properties).
        row : int, default: None
            Tensor row (only used for tensor material properties).
        col : int, default: None
            Tensor column (only used for tensor material properties).
        """
        msg_params = {
            "materialDef": messages.edb_obj_message(self),
            "propertyId": material_property.value,
            "value": messages.value_message(value),
        }
        if component_id is not None:
            msg_params["component"] = messages.edb_internal_id_message(component_id)
        elif row is not None and col is not None:
            msg_params["tensor"] = MaterialDef._tensor_pos_message(col, row)
        self.__stub.SetProperty(pb.MaterialDefSetPropertyMessage(**msg_params))

    def get_property(self, material_property, component_id=None, row=None, col=None):
        """Get a material property value.

            Material properties can be defined in 3 ways:
                1. Simple: A constant value.
                    - To retrieve a simple material property, the ``material_property`` \
                    parameter should be provided.
                2. Anisotropic: A 3x3 tensor consisting of only diagonal entries.
                    - To retrieve a component of an anisotropic material property, the ``material_property`` \
                    and ``component_id`` parameters should be provided.
                    - ``component_id`` specifies the \
                    :term:`anisotropic component ID <Anisotropic Material Property Component IDs>` \
                    of the component to get the value of.
                3. Tensor: A 3x3 tensor consisting of diagonal and off-diagonal entries.
                    - To retrieve an entry in a tensor material property, the ``material_property``, \
                    ``col`` and ``row`` parameters should be provided. The entry at ``T[row,col]`` \
                    will be returned.

        Parameters
        ----------
        material_property : .MaterialProperty
            ID of the material property.
        component_id : int, default: None
            ID of the anisotropic component (only used for anisotropic material properties).
        row : int, default: None
            Tensor row (only used for tensor material properties).
        col : int, default: None
            Tensor column (only used for tensor material properties).

        Returns
        -------
        .Value
        """
        return Value(
            self.__stub.GetProperty(
                MaterialDef._get_property_message(self, material_property, component_id, col, row)
            )
        )

    @property
    def all_properties(self) -> List[MaterialProperty]:
        """:obj:`list` of :class:`.MaterialProperty`: All properties defined in the material definition.

        This property is read-only.
        """
        msg = self.__stub.GetAllProperties(messages.edb_obj_message(self))
        return [MaterialProperty(i) for i in msg.properties]

    def remove_property(self, material_property: MaterialProperty):
        """Remove a property from the material definition.

        Parameters
        ----------
        material_property : .MaterialProperty
            Material property to be removed.
        """
        self.__stub.RemoveProperty(
            MaterialDef._get_property_message(
                messages.edb_obj_message(self), material_property, None, None, None
            )
        )

    def get_dimensions(self, material_property_id: MaterialProperty) -> Tuple[int, int]:
        """Get the dimensions of the tensor of a material property.

            The mappings between the types of material properties and dimensions are as follows:
                - ``Simple`` -> ``1x1``
                - ``Anisotropic`` -> ``3x1``
                - ``Tensor`` -> ``3x3``.

        Parameters
        ----------
        material_property_id : .MaterialProperty
            Material property to get the dimensions of.

        Returns
        -------
        tuple of (int, int)
            The tuple is of the format ``(col, row)``:

            - ``col``: Number of rows of the material property.
            - ``row``: Number of columns of the material property.
        """
        msg = self.__stub.GetDimensions(MaterialDef._property_message(self, material_property_id))
        return msg.tensor.col, msg.tensor.row

    def get_thermal_modifier(
        self, material_property_id: MaterialProperty
    ) -> MaterialPropertyThermalModifier:
        """Get the thermal modifier of a material property.

        Parameters
        ----------
        material_property_id : .MaterialProperty
            Material property to get the thermal modifier of.

        Returns
        -------
        .MaterialPropertyThermalModifier
        """
        return MaterialPropertyThermalModifier(
            self.__stub.GetThermalModifier(
                MaterialDef._property_message(self, material_property_id)
            )
        )

    def set_thermal_modifier(
        self,
        material_property_id: MaterialProperty,
        thermal_modifier: MaterialPropertyThermalModifier,
    ):
        """Set the thermal modifier of the material property.

        Parameters
        ----------
        material_property_id : .MaterialProperty
            Material property to set the thermal modifier on.
        thermal_modifier : .MaterialPropertyThermalModifier
            Thermal modifier to assign to the material property.
        """
        self.__stub.SetThermalModifier(
            pb.SetMaterialDefPropertyMessage(
                materialDef=messages.edb_obj_message(self),
                propertyId=material_property_id.value,
                thermal_modifier=messages.edb_obj_message(thermal_modifier),
            )
        )

    def get_anisotropic_thermal_modifier(self, material_property_id, component_id):
        """Get the thermal modifier of an anisotropic material property.

        Parameters
        ----------
        material_property_id : .MaterialProperty
            Material property to get the thermal modifier of.
        component_id : int
            :term:`Anisotropic component ID <Anisotropic Material Property Component IDs>` of the \
            material property to get the thermal modifier of.

        Returns
        -------
        .MaterialPropertyThermalModifier
        """
        return MaterialPropertyThermalModifier(
            self.__stub.GetAnisotropicThermalModifier(
                pb.MaterialDefPropertyComponentMessage(
                    materialDef=messages.edb_obj_message(self),
                    propertyId=material_property_id.value,
                    component=component_id,
                )
            )
        )

    def set_anisotropic_thermal_modifier(
        self, material_property_id, component_id, thermal_modifier
    ):
        """Set the thermal modifier of an anisotropic material property.

        Parameters
        ----------
        material_property_id : .MaterialProperty
            Type of material property to set the thermal modifier on.
        component_id : int
            :term:`Anisotropic component ID <Anisotropic Material Property Component IDs>` of the \
            material property to set the thermal modifier on.
        thermal_modifier : .MaterialPropertyThermalModifier
        """
        self.__stub.SetAnisotropicThermalModifier(
            pb.SetMaterialDefPropertyComponentMessage(
                materialDef=messages.edb_obj_message(self),
                propertyId=material_property_id.value,
                component=component_id,
                thermal_modifier=messages.edb_obj_message(thermal_modifier),
            )
        )

    @staticmethod
    def _tensor_pos_message(col, row):
        return pb.TensorPositionMessage(
            col=col,
            row=row,
        )

    @staticmethod
    def _get_property_message(material_def, material_property, component, col, row):
        msg_params = {
            "materialDef": messages.edb_obj_message(material_def),
            "propertyId": material_property.value,
        }
        if component is not None:
            msg_params["component"] = messages.edb_internal_id_message(component)
        elif row is not None and col is not None:
            msg_params["tensor"] = MaterialDef._tensor_pos_message(col, row)
        return pb.MaterialDefGetPropertyMessage(**msg_params)

    @staticmethod
    def _property_message(material_def, material_property_id):
        return pb.MaterialDefPropertyMessage(
            materialDef=messages.edb_obj_message(material_def),
            propertyId=material_property_id.value,
        )
